import os

try:
    import jax
except Exception:  # pragma: no cover
    jax = None


def gpu_tests_enabled() -> bool:
    return os.environ.get("MICROJAX_GPU_TESTS", "0") in {"1", "true", "True"}


def has_cuda() -> bool:
    if jax is None:
        return False
    try:
        return len(jax.devices("cuda")) > 0
    except Exception:
        return False


def is_a100() -> bool:
    if not has_cuda():
        return False
    try:
        return any("A100" in d.device_kind for d in jax.devices("cuda"))
    except Exception:
        return False




# This file is a modified and extended version of code from the `caustics` package:
#   https://github.com/fbartolic/caustics
# Originally developed by Fran Bartolic under the MIT License.
#
# modifications and extensions have been made by Shota Miyazaki for the `microjax` project.
#
# SPDX-FileCopyrightText: 2022 Fran Bartolic
# SPDX-FileCopyrightText: 2025 Shota Miyazaki
# SPDX-License-Identifier: MIT

__all__ = [
    "magnifications",
]

from functools import partial

import jax.numpy as jnp
from jax import jit, lax 

from .extended_source import mag_extended_source
from ..point_source import _images_point_source

from ..multipole import _mag_hexadecapole

from ..utils import *


@partial(jit, static_argnames=("nlenses"))
def _caustics_proximity_test(
    w, z, z_mask, rho, delta_mu_multi, nlenses=2, c_m=1e-02, gamma=0.02, c_f=4., rho_min=1e-03, **params
):
    if nlenses == 2:
        a, e1 = params["a"], params["e1"]
        e2 = 1.0 - e1
        # Derivatives
        f = lambda z: - e1 / (z - a) - e2 / (z + a)
        f_p = lambda z: e1 / (z - a) ** 2 + e2 / (z + a) ** 2
        f_pp = lambda z: 2 * (e1 / (a - z) ** 3 - e2 / (a + z) ** 3)

    elif nlenses == 3:
        a, r3, e1, e2 = params["a"], params["r3"], params["e1"], params["e2"]
        # Derivatives
        f = lambda z: - e1 / (z - a) - e2 / (z + a) - (1 - e1 - e2) / (z + r3)
        f_p = (
            lambda z: e1 / (z - a) ** 2
            + e2 / (z + a) ** 2
            + (1 - e1 - e2) / (z + r3) ** 2
        )
        f_pp = (
            lambda z: 2 * (e1 / (a - z) ** 3 - e2 / (a + z) ** 3)
            + (1 - e1 - e2) / (z + r3) ** 3
        )
    zbar = jnp.conjugate(z)
    zhat = jnp.conjugate(w) - f(z)

    # Derivatives
    fp_z     = f_p(z)
    fpp_z    = f_pp(z)
    fp_zbar  = f_p(zbar)
    fp_zhat  = f_p(zhat)
    fpp_zbar = f_pp(zbar)
    J        = 1.0 - jnp.abs(fp_z * fp_zbar)

    # Multipole test and cusp test
    mu_cusp = 6 * jnp.imag(3 * fp_zbar**3.0 * fpp_z**2.0) / J**5 * (rho + rho_min)**2
    mu_cusp = jnp.sum(jnp.abs(mu_cusp) * z_mask, axis=0)
    test_multipole_and_cusp = gamma * mu_cusp + delta_mu_multi < c_m

    # False images test
    Jhat = 1 - jnp.abs(fp_z * fp_zhat)
    factor = jnp.abs(J * Jhat**2 / 
                     (Jhat*fpp_zbar*fp_z - jnp.conjugate(Jhat)  * fpp_z * fp_zbar * fp_zhat)
                     )
    test_false_images = 0.5 * (~z_mask * factor).sum(axis=0) > c_f * (rho + rho_min)
    test_false_images = jnp.where((~z_mask).sum(axis=0)==0, 
                                  jnp.ones_like(test_false_images, dtype=jnp.bool_), 
                                  test_false_images
                                  )
    return test_false_images & test_multipole_and_cusp


def _planetary_caustic_test(w, rho, c_p=2., **params):
    e1, a = params["e1"], params["a"]
    s = 2 * a
    q = e1 / (1.0 - e1)
    x_cm = (2*e1 - 1)*a
    w_pc = -1/s 
    delta_pc = 3*jnp.sqrt(q)/s
    return (w_pc - w).real**2 + (w_pc - w).imag**2 > c_p*(rho**2 + delta_pc**2)


@partial(
    jit,
    static_argnames=(
        "nlenses",
        "npts_limb",
        "limb_darkening",
        "npts_ld",
    ),
)
def magnifications(
    w_points,
    rho,
    nlenses=2,
    npts_limb=200,
    limb_darkening=False,
    u1=0.0,
    npts_ld=100,
    **params
):
    """
    Compute the extended source magnification for a system with `nlenses` lenses 
    and a source star radius `rho` at a set of complex points `w_points` in the 
    source plane. This function calls either [`caustics.mag_hexadecapole`][] or 
    [`caustics.mag_extended_source`][] at each point in `w_points` depending on
    whether or not the hexadecapole approximation is accurate enough at that point. 

    If `nlenses` is 2 (binary lens) or 3 (triple lens), the coordinate system is
    set up such that the the origin is at the center of mass of the first two 
    lenses which are both located on the real line. The location of the first 
    lens is $-sq/(1 + q)$ and the second lens is at $s/(1 + q)$. The optional 
    third lens is located at an arbitrary position in the complex plane 
    $r_3e^{-i\psi}$. The magnification is computed using contour integration in
    the image plane. Boolean flag `limb_darkening` indicates whether linear 
    limb-darkening needs to taken into account. If `limb_darkening` is set to 
    `True` the linear limb-darkening coefficient `u1` needs to be specified as 
    well. Note that turning on this flag slows down the computation by up to an 
    order of magnitude.

    If `nlenses` is 2 only the parameters `s` and `q` should be specified. If 
    `nlenses` is 3, the parameters `s`, `q`, `q3`, `r3` and `psi` should be 
    specified.

    !!! note

        Turning on limb-darkening (`limb_darkening=True`) slows down the 
        computation by up to an order of magnitude.
    
    !!! warning

        At the moment the test determining whether or not to use the hexadecapole
        approximation does not work for triple lenses so the function will use
        full contour integration at every point. This substantially slows down
        the computation. See https://github.com/fbartolic/caustics/issues/19.

    Args:
        w_points (array_like): Source positions in the complex plane.
        rho (float): Source radius in Einstein radii.
        nlenses (int): Number of lenses in the system.
        npts_limb (int, optional): Initial number of points uniformly distributed
            on the source limb when computing the point source magnification.
            The final number of points is greater than this value because
            the number of points is decreased geometrically by a factor of
            1/2 until it reaches 2.
        limb_darkening (bool, optional): If True, compute the magnification of
            a limb-darkened source. If limb_darkening is enabled the u1 linear
            limb-darkening coefficient needs to be specified. Defaults to False.
        u1 (float, optional): Linear limb darkening coefficient. Defaults to 0..
        npts_ld (int, optional): Number of points at which the stellar brightness
            function is evaluated when computing contour integrals 
            $\int P(z_1^\prime, z_2) dz_1^\prime$ and 
            $\int Q(z_1, z_2^\prime) dz_2^\prime$ (see Dominik 1998). Defaults 
            to 100.
         s (float): Separation between the two lenses. The first lens is located 
            at $-sq/(1 + q)$ and the second lens is at $s/(1 + q)$ on the real line.
        q (float): Mass ratio defined as $m_2/m_1$.
        q3 (float): Mass ratio defined as $m_3/m_1$.
        r3 (float): Magnitude of the complex position of the third lens.
        psi (float): Phase angle of the complex position of the third lens.
        roots_itmax (int, optional): Number of iterations for the root solver.
        roots_compensated (bool, optional): Whether to use the compensated
            arithmetic version of the Ehrlich-Aberth root solver.

    Returns:
        array_like: Magnification array.
    """
    if nlenses == 1:
        _params = {}
        x_cm = 0 # miyazaki
    elif nlenses == 2:
        s, q = params["s"], params["q"]
        a = 0.5 * s
        e1 = q / (1.0 + q) 
        _params = {"a": a, "e1": e1}
        x_cm = a*(1 - q)/(1 + q)

    # Trigger the full calculation everywhere because I haven't figured out 
    # how to implement the ghost image test for nlenses > 2 yet
    elif nlenses == 3:
        s, q, q3, r3, psi = params["s"], params["q"], params["q3"], params["r3"], params["psi"]
        a = 0.5 * s
        e1 = q / (1.0 + q + q3)
        e2 = 1.0 / (1.0 + q + q3) #miyazaki
        r3 = r3*jnp.exp(1j * psi)
        _params = {"a": a, "r3": r3, "e1": e1, "e2": e2}
        x_cm = a * (1.0 - q) / (1.0 + q)

    else:
        raise ValueError("nlenses must be <= 3")


    # Compute point images for a point source
    z, z_mask = _images_point_source(
        w_points - x_cm, #miyazaki
        #w_points + x_cm,
        nlenses=nlenses,
        **_params
    )

    if nlenses==1:
        test = w_points > 2*rho
        mu_multi, delta_mu_multi = _mag_hexadecapole(z, z_mask, rho, nlenses=nlenses, **_params) #miyazaki
    elif nlenses==2:
        # Compute hexadecapole approximation at every point and a test where it is
        # sufficient
        mu_multi, delta_mu_multi = _mag_hexadecapole(z, z_mask, rho, nlenses=nlenses, **_params)
        test1 = _caustics_proximity_test(
            w_points - x_cm, z, z_mask, rho, delta_mu_multi, nlenses=nlenses,  **_params #miyazaki
            #w_points + x_cm, z, z_mask, rho, delta_mu_multi, nlenses=nlenses,  **_params
        )
        test2 = _planetary_caustic_test(w_points - x_cm, rho, **_params)
        #test2 = _planetary_caustic_test(w_points + x_cm, rho, **_params)

        test = lax.cond(
            q < 0.01, 
            lambda:test1 & test2,
            lambda:test1,
        )
    elif nlenses == 3:
        test = jnp.zeros_like(w_points).astype(jnp.bool_)
    
    mag_full = lambda w: mag_extended_source(
        w,
        rho,
        nlenses=nlenses,
        npts_limb=npts_limb,
        limb_darkening=limb_darkening,
        u1=u1,
        npts_ld=npts_ld,
        **params,
    )

    # Iterate over w_points and execute either the hexadecapole  approximation
    # or the full extended source calculation. `vmap` cannot be used here because
    # `lax.cond` executes both branches within vmap.
    return lax.map(
        lambda xs: lax.cond(
            xs[0],
            lambda _: xs[1],
            mag_full,
            xs[2],
        ),
        [test, mu_multi, w_points],
        #        jnp.stack([mask_test, mu_approx,  w_points]).T,
    )



"""Likelihood and simple linear fits for microlensing photometry.

This module provides small, JAX-friendly utilities to compute weighted linear
least-squares fits and negative log-likelihoods (NLL) under Gaussian models.
Two NLL variants are implemented:

- ``nll_ulens``: fast path with diagonal observational noise and diagonal
  Gaussian priors on source and blend fluxes.
- ``nll_ulens_general``: full-rank observational covariance and arbitrary
  Gaussian priors (mean and covariance).

All routines operate on ``jnp.ndarray`` inputs and are differentiable, making
them suitable for gradient-based optimization or inference.
"""

from typing import Tuple, Union

import jax.numpy as jnp
from jax.scipy.linalg import solve

# Consistent array alias used across modules
Array = jnp.ndarray

def linear_chi2(x: Array, y: Array, err: Union[float, Array] = 0.0) -> Tuple[float, float, float, float, float]:
    """Weighted linear fit ``y ≈ a + b x`` and chi-squared.

    Performs a closed-form, weighted least-squares fit with independent
    Gaussian errors. If any entry of ``err`` is 0 (default), that point is
    treated as unweighted with weight 1.0.

    Parameters
    ----------
    x : jnp.ndarray, shape (n,)
        Predictor values.
    y : jnp.ndarray, shape (n,)
        Observed responses.
    err : float or jnp.ndarray, shape (n,), optional
        Per-point standard deviation. A scalar broadcasts to all points. Zeros
        are replaced by 1.0 in the weights.

    Returns
    -------
    b : float
        Best-fit slope.
    be : float
        Standard error on the slope.
    a : float
        Best-fit intercept.
    ae : float
        Standard error on the intercept.
    chi2 : float
        Chi-squared of the fit evaluated at the best-fit parameters.

    Notes
    -----
    The solution minimizes ``sum_i w_i (y_i - a - b x_i)^2`` with
    ``w_i = 1/err_i^2`` when ``err_i > 0`` else ``w_i = 1``.
    """
    wt = jnp.where(err > 0, 1.0 / (err ** 2), 1.0)
    sumw = jnp.sum(wt)
    sumx = jnp.sum(wt * x)
    sumy = jnp.sum(wt * y)
    sumxx = jnp.sum(wt * x * x)
    sumxy = jnp.sum(wt * x * y)
    det = sumw * sumxx - sumx * sumx

    a = (sumxx * sumy - sumx * sumxy) / det
    b = (sumw * sumxy - sumx * sumy) / det

    residual = y - a - b * x
    chi2 = jnp.sum(wt * residual ** 2)
    ae = jnp.sqrt(sumxx / det)
    be = jnp.sqrt(sumw / det)
    return b, be, a, ae, chi2


def nll_ulens(
    flux: Array,
    M: Array,
    sigma2_obs: Array,
    sigma2_fs: float,
    sigma2_fb: float,
) -> float:
    """Negative log-likelihood with diagonal noise and diagonal priors.

    Assumes a linear flux model ``flux ≈ M @ [fs, fb]`` with independent
    observational errors ``C = diag(sigma2_obs)`` and independent Gaussian
    priors on ``fs`` and ``fb`` with variances ``sigma2_fs`` and ``sigma2_fb``.
    All integrals over ``fs, fb`` are done analytically, yielding the standard
    marginalized Gaussian NLL.

    Parameters
    ----------
    flux : jnp.ndarray, shape (n,)
        Observed fluxes.
    M : jnp.ndarray, shape (n, 2)
        Design matrix with columns ``[m_fs, 1]`` where ``m_fs`` is the source
        model term and the constant column captures blend flux.
    sigma2_obs : jnp.ndarray, shape (n,)
        Observational variances (diagonal of the noise covariance).
    sigma2_fs : float
        Prior variance for source flux ``fs``.
    sigma2_fb : float
        Prior variance for blend flux ``fb``.

    Returns
    -------
    nll : float
        Negative log-likelihood value.

    Notes
    -----
    The expression equals 0.5 * (mahalanobis + log|C| + log|Lambda| + log|A|)
    where ``A = M^T C^{-1} M + Lambda^{-1}`` and ``Lambda`` is diagonal with
    entries ``[sigma2_fs, sigma2_fb]``.
    """
    # Prior precision matrix Lambda^{-1} = diag(1/sigma2_fs, 1/sigma2_fb)
    lambda_fs = 1.0 / sigma2_fs
    lambda_fb = 1.0 / sigma2_fb

    # Compute C^{-1} @ M, using element-wise division since C is diagonal
    Cinv_M = M / sigma2_obs[:, None]  # shape (n, 2)

    # Compute matrix A = M^T @ C^{-1} @ M + Lambda^{-1}
    # A is 2x2 symmetric matrix
    a11 = lambda_fs + jnp.dot(M[:, 0], Cinv_M[:, 0])
    a22 = lambda_fb + jnp.dot(M[:, 1], Cinv_M[:, 1])
    a12 = jnp.dot(M[:, 0], Cinv_M[:, 1])  # same as a21

    # Compute vector b = M^T @ C^{-1} @ flux
    Cinv_flux = flux / sigma2_obs
    b1 = jnp.dot(M[:, 0], Cinv_flux)
    b2 = jnp.dot(M[:, 1], Cinv_flux)

    # Solve A @ mu = b for posterior mean mu = (mu_fs, mu_fb)
    detA = a11 * a22 - a12 * a12
    invA00 =  a22 / detA
    invA11 =  a11 / detA
    invA01 = -a12 / detA
    mu_fs = invA00 * b1 + invA01 * b2
    mu_fb = invA01 * b1 + invA11 * b2

    # Compute Mahalanobis part: flux^T @ C^{-1} @ flux - b^T @ A^{-1} @ b
    # Compute log-determinant terms
    mahalanobis = jnp.dot(flux, Cinv_flux) - (mu_fs * b1 + mu_fb * b2)
    logdet_C = jnp.sum(jnp.log(sigma2_obs))                     # log|C|
    logdet_Lambda = jnp.log(sigma2_fs) + jnp.log(sigma2_fb)     # log|Lambda|
    logdet_A = jnp.log(detA)                                    # log|A|

    # Final NLL value
    nll = 0.5 * (mahalanobis + logdet_C + logdet_Lambda + logdet_A)
    return nll
    
def nll_ulens_general(
    flux: Array,
    M: Array,
    C: Array,
    mu: Array,
    Lambda: Array,
) -> float:
    """Negative log-likelihood with full-rank noise and Gaussian priors.

    Generalizes :func:`nll_ulens` to non-diagonal observational covariance
    ``C`` and an arbitrary Gaussian prior ``N(mu, Lambda)`` over ``[fs, fb]``.
    The result is the marginalized NLL after integrating out the linear
    parameters analytically.

    Parameters
    ----------
    flux : jnp.ndarray, shape (n,)
        Observed fluxes.
    M : jnp.ndarray, shape (n, 2)
        Design matrix mapping ``[fs, fb]`` to the model flux.
    C : jnp.ndarray, shape (n, n)
        Observational covariance matrix (symmetric positive definite).
    mu : jnp.ndarray, shape (2,)
        Prior mean for ``[fs, fb]``.
    Lambda : jnp.ndarray, shape (2, 2)
        Prior covariance for ``[fs, fb]`` (symmetric positive definite).

    Returns
    -------
    nll : float
        Negative log-likelihood value.

    Notes
    -----
    Defines ``A = Lambda^{-1} + M^T C^{-1} M`` and posterior mean
    ``m = A^{-1} M^T C^{-1} (flux - M mu)``. The NLL equals
    ``0.5 * ( (flux - M mu)^T C^{-1} (flux - M mu) - m^T A m + log|C| + log|Lambda| + log|A| )``.
    """
    # Inverse of prior covariance
    Lambda_inv = jnp.linalg.inv(Lambda)            # shape (2, 2)

    # Solve C⁻¹ @ M efficiently
    Cinv_M = solve(C, M)                            # shape (n, 2)
    Mt_Cinv_M = M.T @ Cinv_M                        # shape (2, 2)

    # Posterior precision matrix
    A = Lambda_inv + Mt_Cinv_M                      # shape (2, 2)

    # Residual vector r = flux - M @ mu
    r = flux - M @ mu                               # shape (n,)

    # Solve C⁻¹ @ r efficiently
    Cinv_r = solve(C, r)                            # shape (n,)
    Mt_Cinv_r = M.T @ Cinv_r                        # shape (2,)

    # Posterior mean (optional, not returned but used for Mahalanobis term)
    posterior_mean = jnp.linalg.solve(A, Mt_Cinv_r) # shape (2,)

    # Mahalanobis term
    mahal_term = jnp.dot(r, Cinv_r) - Mt_Cinv_r @ posterior_mean

    # Log-determinant terms
    _, logdet_C = jnp.linalg.slogdet(C)
    _, logdet_Lambda = jnp.linalg.slogdet(Lambda)
    _, logdet_A = jnp.linalg.slogdet(A)

    # Final negative log-likelihood
    nll = 0.5 * (mahal_term + logdet_C + logdet_Lambda + logdet_A)
    return nll



# This file is a modified and extended version of code from the `caustics` package:
#   https://github.com/fbartolic/caustics
# Originally developed by Fran Bartolic under the MIT License.
#
# modifications and extensions have been made by Shota Miyazaki for the `microjax` project.
#
# SPDX-FileCopyrightText: 2022 Fran Bartolic
# SPDX-FileCopyrightText: 2025 Shota Miyazaki
# SPDX-License-Identifier: MIT


from functools import partial

import numpy as np
import jax.numpy as jnp
from jax import jit

from ..utils import *

from ..point_source import lens_eq
from ..utils import trapz_zero_avoiding


def _integrate_gauss_legendre(f, a, b, n=100):
    pts, weights = np.polynomial.legendre.leggauss(n)
    pts_rescaled = 0.5 * (b - a) * pts[:, None] + 0.5 * (b + a)
    return jnp.sum(0.5 * (b - a) * f(pts_rescaled) * weights[:, None], axis=0)

def _integrate_unif(z, tidx):
    # Trapezoidal rule
    I1 = trapz_zero_avoiding(0.5 * z.real, z.imag, tidx)
    I2 = trapz_zero_avoiding(-0.5 * z.imag, z.real, tidx)
    return I1 + I2

def _brightness_profile(z, rho, w0, u1=0.0, nlenses=2, **params):
    w = lens_eq(z, nlenses=nlenses, **params)
    r = jnp.abs(w - w0) / rho

    def safe_for_grad_sqrt(x):
        return jnp.sqrt(jnp.where(x > 0.0, x, 0.0))

    # See Dominik 1998 for details
    B_r = jnp.where(
        r <= 1.0,
        1 + safe_for_grad_sqrt(1 - r**2),
        1 - safe_for_grad_sqrt(1 - 1.0 / r**2),
    )

    I = 3.0 / (3.0 - u1) * (u1 * B_r + 1.0 - 2.0 * u1)
    return I


@partial(jit, static_argnames=("nlenses", "npts"))
def _integrate_ld(z, tidx, w0, rho, u1=0.0, nlenses=2, npts=100, **params):
    def P(_, y0, xl, yl):
        # Construct grid in z2 and evaluate the brightness profile at each point
        a, b = y0 * jnp.ones_like(xl), yl  # lower and upper limits
        abs_delta = jnp.abs(b - a)

        # Split each integral into two intervals with the same number of points
        # and integrate each using Gauss Legendre quadrature
        mask = b > a
        split_points = jnp.where(
            mask,
            b - 2 * rho,
            b + 2 * rho,
        )
        split_points = jnp.where(
            0.5 * abs_delta <= 2 * rho, a + 0.5 * abs_delta, split_points
        )

        # First interval
        f = lambda y: _brightness_profile(
            xl + 1j * y, rho, w0, u1=u1, nlenses=nlenses, **params
        )
        npts1 = int(npts / 2)
        npts2 = npts - npts1
        I1 = _integrate_gauss_legendre(f, a, split_points, n=npts1)

        # Second interval
        I2 = _integrate_gauss_legendre(f, split_points, b, n=npts2)
        I = I1 + I2

        return -0.5 * I

    def Q(x0, _, xl, yl):
        # Construct grid in z1 and evaluate the brightness profile at each point
        a, b = x0 * jnp.ones_like(yl), xl
        abs_delta = jnp.abs(b - a)

        # Split each integral into two intervals with the same number of points
        # and integrate each using Gauss Legendre quadrature
        mask = b > a
        split_points = jnp.where(
            mask,
            b - 2 * rho,
            b + 2 * rho,
        )
        split_points = jnp.where(
            0.5 * abs_delta <= 2 * rho, a + 0.5 * abs_delta, split_points
        )

        # First interval
        f = lambda x: _brightness_profile(
            x + 1j * yl, rho, w0, u1=u1, nlenses=nlenses, **params
        )
        npts1 = int(npts / 2)
        npts2 = npts - npts1
        I1 = _integrate_gauss_legendre(f, a, split_points, n=npts1)

        # Second interval
        I2 = _integrate_gauss_legendre(f, split_points, b, n=npts2)
        I = I1 + I2

        return 0.5 * I

    # Centroid of the contour
    z0 = z.sum() / (tidx + 1)

    # Evaluate the P and Q integrals using Trapezoidal rule
    _P = P(z0.real, z0.imag, z.real, z.imag)
    _Q = Q(z0.real, z0.imag, z.real, z.imag)

    I1 = trapz_zero_avoiding(_P, z.real, tidx)
    I2 = trapz_zero_avoiding(_Q, z.imag, tidx)

    return I1 + I2



"""Inverse-ray finite-source magnification integrators.

This module provides polar-grid, inverse-ray integrators for extended sources
subject to gravitational microlensing by multiple lenses. Two brightness
profiles are implemented:

- ``mag_uniform``: uniform surface brightness disk.
- ``mag_limb_dark``: linear limb-darkened disk.

The integration domain in image space is seeded using the mapped source limb
(`calc_source_limb`) and partitioned into radial/azimuthal regions
(`define_regions`) to concentrate samples near caustics. Boundary crossings in
angle are handled by a numerically stable 4-point cubic interpolation
(`cubic_interp`), with optional smoothing factors to reduce aliasing at the
source limb.

Notes
- Both routines shift coordinates to the center-of-mass frame for improved
  stability and consistency with point-source helpers.
- ``lax.scan`` over regions reduces peak memory usage relative to a full
  vectorized ``vmap`` across all regions.
- Increase ``bins_*`` and resolutions when approaching caustics or for larger
  sources to improve accuracy (with corresponding runtime/memory costs).
"""

import jax 
import jax.numpy as jnp
from jax import jit, lax, vmap, custom_jvp
from functools import partial
from microjax.point_source import lens_eq, _images_point_source
from microjax.inverse_ray.merge_area import calc_source_limb, define_regions
from microjax.inverse_ray.limb_darkening import Is_limb_1st
from microjax.inverse_ray.boundary import in_source, distance_from_source, calc_facB
from typing import Mapping, Sequence, Tuple, Callable, Optional, Union

# Simple alias for readability in type hints
Array = jnp.ndarray

#@partial(jit, static_argnames=("nlenses", "cubic", "r_resolution", "th_resolution", "Nlimb", "u1",
#                               "offset_r", "offset_th", "delta_c"))
def mag_limb_dark(
    w_center: complex,
    rho: float,
    nlenses: int = 2,
    u1: float = 0.0,
    r_resolution: int = 500,
    th_resolution: int = 500,
    Nlimb: int = 500,
    bins_r: int = 50,
    bins_th: int = 120,
    margin_r: float = 0.5,
    margin_th: float = 0.5,
    delta_c: float = 0.01,
    **_params: float,
) -> Array:
    """Compute finite-source magnification with linear limb darkening.

    This routine evaluates the magnification of a circular source centered at
    `w_center` with radius `rho` using a polar grid integration in image space
    and an inverse ray approach. It supports binary (``nlenses=2``) and a
    specific triple-lens configuration (``nlenses=3``). The surface brightness
    profile on the source is linear limb-darkened:

        I(r)/I0 = 1 - u1 * (1 - sqrt(1 - (r/rho)^2)),

    implemented via ``Is_limb_1st``. The image-space region to integrate is
    constructed from points on the mapped source limb using
    ``calc_source_limb`` and partitioned by ``define_regions`` for better
    conditioning around caustics. Angular boundary crossings are located with a
    stable 4-point cubic (Lagrange) interpolant (``cubic_interp``). A smooth
    transition factor ``calc_facB`` controlled by ``delta_c`` reduces aliasing
    at the limb.

    Parameters
    - w_center: complex – Source center in lens plane units (Einstein radius).
    - rho: float – Source radius (same units as `w_center`).
    - nlenses: int – Number of lenses (2 supported; 3 supported for provided
      params).
    - u1: float – Linear limb-darkening coefficient in [0, 1].
    - r_resolution: int – Number of radial samples per region.
    - th_resolution: int – Number of angular samples per region.
    - Nlimb: int – Number of samples on the source limb used to seed regions.
    - bins_r: int – Number of radial bins to split regions.
    - bins_th: int – Number of angular bins to split regions.
    - margin_r: float – Extra radial margin added to each bin (in units of
      `rho`).
    - margin_th: float – Extra angular margin added to each bin (radians).
    - delta_c: float – Smoothing scale for boundary contribution factor
      ``calc_facB``; smaller sharpens boundary, larger smooths.
    - **_params: Mapping of lens parameters depending on ``nlenses``.
      For nlenses=2 expect ``q`` and ``s``; for nlenses=3 expect ``s``, ``q``,
      ``q3``, ``r3``, ``psi``.

    Returns
    - magnification: float – Limb-darkened finite-source magnification.

    Notes
    - Coordinates are internally shifted by the center-of-mass offset for
      numerical stability and consistency with point-source helpers.
    - The result is normalized by ``rho**2`` (no factor of π) because the limb
      darkening weights are included explicitly in the integrand.
    - For large ``rho`` or near-caustic configurations, increase ``bins_*`` and
      ``*_resolution`` to improve accuracy at the cost of memory/runtime.
    """
    if nlenses == 2:
        q, s = _params["q"], _params["s"]
        a  = 0.5 * s
        e1 = q / (1.0 + q)
        _params = {"q": q, "s": s, "a": a, "e1": e1}
    elif nlenses == 3:
        s, q, q3, r3, psi = _params["s"], _params["q"], _params["q3"], _params["r3"], _params["psi"]
        a = 0.5 * s
        total_mass = 1.0 + q + q3
        e1 = q / total_mass
        e2 = 1.0 / total_mass 
        _params = {"a": a, "r3": r3, "e1": e1, "e2": e2, "q": q, "s": s, "q3": q3, "psi": psi}
    
    shifted = 0.5 * s * (1 - q) / (1 + q)  
    w_center_shifted = w_center - shifted
    image_limb, mask_limb = calc_source_limb(w_center, rho, Nlimb, **_params)
    r_scan, th_scan = define_regions(image_limb, mask_limb, rho, bins_r=bins_r, bins_th=bins_th, 
                                     margin_r=margin_r, margin_th=margin_th, nlenses=nlenses)

    def _process_r(r0: float, th_values: Array) -> Array:
        """Integrand over angle for a fixed radius.

        Computes the limb-darkened contribution for radius ``r0`` by sampling
        angles ``th_values``. It classifies samples as inside/outside the source
        disk via ``in_source(distance_from_source(...))`` and adds smoothed
        boundary terms using ``cubic_interp`` and ``calc_facB``.

        Returns the summed area contribution (not yet multiplied by ``dr``).
        """
        dth = (th_values[1] - th_values[0])
        distances = distance_from_source(r0, th_values, w_center_shifted, shifted, nlenses=nlenses, **_params)
        in_num = in_source(distances, rho)
        Is     = Is_limb_1st(distances / rho, u1=u1)
        zero_term = 1e-10
        in0_num, in1_num, in2_num, in3_num, in4_num = in_num[:-4], in_num[1:-3], in_num[2:-2], in_num[3:-1], in_num[4:]
        d0, d1, d2, d3, d4 = distances[:-4], distances[1:-3], distances[2:-2], distances[3:-1], distances[4:]
        th0, th1, th2, th3 = jnp.arange(4)
        num_inside  = in1_num * in2_num * in3_num
        num_B1      = (1.0 - in1_num) * in2_num * in3_num
        num_B2      = in1_num * in2_num * (1.0 - in3_num)
        th_est_B1   = cubic_interp(rho, d0, d1, d2, d3, th0, th1, th2, th3, epsilon=zero_term)
        th_est_B2   = cubic_interp(rho, d1, d2, d3, d4, th0, th1, th2, th3, epsilon=zero_term)
        delta_B1    = jnp.clip(th2 - th_est_B1, 0.0, 1.0) + zero_term
        delta_B2    = jnp.clip(th_est_B2 - th1, 0.0, 1.0) + zero_term
        fac_B1 = calc_facB(delta_B1, delta_c)
        fac_B2 = calc_facB(delta_B2, delta_c)
        area_inside = r0 * dth * Is[2:-2] * num_inside
        area_B1     = r0 * dth * Is[2:-2] * fac_B1 * num_B1
        area_B2     = r0 * dth * Is[2:-2] * fac_B2 * num_B2
        return jnp.sum(area_inside + area_B1 + area_B2)

    #@jax.checkpoint 
    def _compute_for_range(r_range: Array, th_range: Array) -> Array:
        """Integrate over a rectangular image-space subregion.

        - r_range: length-2 array giving [r_min, r_max].
        - th_range: length-2 array giving [theta_min, theta_max].

        Builds uniform 1D grids of sizes ``r_resolution`` and ``th_resolution``
        and performs a rectangle-rule accumulation over radius with per-radius
        angular sums from ``_process_r``. Returns the total area contribution
        of this subregion.
        """
        r_values = jnp.linspace(r_range[0], r_range[1], r_resolution, endpoint=True)
        th_values = jnp.linspace(th_range[0], th_range[1], th_resolution, endpoint=True)
        area_r = vmap(lambda r: _process_r(r, th_values))(r_values)
        dr = r_values[1] - r_values[0]
        total_area = dr * jnp.sum(area_r) # trapezoidal integration
        return total_area
    
    inputs = (r_scan, th_scan)
    if(1): # memory efficient but seems complex implementation for jax.checkpoint.
        #@jax.checkpoint
        def scan_images(carry, inputs):
            r_range, th_range = inputs
            total_area = _compute_for_range(r_range, th_range)
            return carry + total_area, None
        magnification_unnorm, _ = lax.scan(scan_images, 0.0, inputs, unroll=1)
    if(0): # vmap case. subtle improvement in speed but worse in memory. More careful for chunking size.
        total_areas = vmap(_compute_for_range, in_axes=(0, 0))(r_scan, th_scan)
        magnification_unnorm = jnp.sum(total_areas)
    magnification = magnification_unnorm / rho**2 
    return magnification 

#@partial(jit, static_argnames=("nlenses", "r_resolution", "th_resolution", "Nlimb", "offset_r", "offset_th", "cubic",))
def mag_uniform(
    w_center: complex,
    rho: float,
    nlenses: int = 2,
    r_resolution: int = 500,
    th_resolution: int = 500,
    Nlimb: int = 500,
    bins_r: int = 50,
    bins_th: int = 120,
    margin_r: float = 0.5,
    margin_th: float = 0.5,
    **_params: float,
) -> Array:
    """Compute finite-source magnification for a uniform-brightness disk.

    Uses the same region construction and polar-grid integration strategy as
    ``mag_limb_dark`` but with a uniform surface brightness profile. The
    integrand is the area fraction inside the source with sub-cell angular
    crossing handled by a stable cubic interpolation in angle.

    Parameters
    - w_center: complex – Source center in Einstein-radius units.
    - rho: float – Source radius.
    - nlenses: int – Number of lenses (2 supported; 3 supported for provided
      params).
    - r_resolution: int – Number of radial samples per region.
    - th_resolution: int – Number of angular samples per region.
    - Nlimb: int – Number of samples on the source limb used to seed regions.
    - bins_r: int – Number of radial bins for region partitioning.
    - bins_th: int – Number of angular bins for region partitioning.
    - margin_r: float – Extra radial margin per bin (in units of ``rho``).
    - margin_th: float – Extra angular margin per bin (radians).
    - **_params: Lens parameters depending on ``nlenses`` (same as in
      ``mag_limb_dark``).

    Returns
    - magnification: float – Uniform finite-source magnification normalized by
      ``rho**2 * pi``.

    Notes
    - Internally shifts coordinates by the lens center-of-mass offset.
    - Region-wise ``lax.scan`` is default for better peak-memory behavior; a
      fully vectorized alternative via ``vmap`` is available but uses more
      memory.
    - Sensitivity near caustics can be improved by increasing ``bins_*`` and
      ``*_resolution`` or broadening ``margin_*``.
    """
    
    if nlenses == 2:
        q, s = _params["q"], _params["s"]
        a  = 0.5 * s
        e1 = q / (1.0 + q)
        _params = {"q": q, "s": s, "a": a, "e1": e1}
    elif nlenses == 3:
        s, q, q3, r3, psi = _params["s"], _params["q"], _params["q3"], _params["r3"], _params["psi"]
        a = 0.5 * s
        total_mass = 1.0 + q + q3
        e1 = q / total_mass
        e2 = 1.0 / total_mass 
        _params = {"a": a, "r3": r3, "e1": e1, "e2": e2, "q": q, "s": s, "q3": q3, "psi": psi}
    
    shifted = a * (1.0 - q) / (1.0 + q)  
    w_center_shifted = w_center - shifted
    image_limb, mask_limb = calc_source_limb(w_center, rho, Nlimb, **_params, nlenses=nlenses)
    r_scan, th_scan = define_regions(image_limb, mask_limb, rho, bins_r=bins_r, bins_th=bins_th, 
                                     margin_r=margin_r, margin_th=margin_th, nlenses=nlenses)

    #@jax.checkpoint 
    def _process_r(r0: float, th_values: Array) -> Array:
        """Angular accumulation at fixed radius for a uniform source.

        Classifies points as inside/outside the source and corrects the two
        nearest angular cells that cross the source limb using a cubic estimate
        of the crossing angle. Returns the summed (angular) area at radius
        ``r0`` (prior to multiplying by ``dr``).
        """
        dth = (th_values[1] - th_values[0])
        distances = distance_from_source(r0, th_values, w_center_shifted, shifted, nlenses=nlenses, **_params)
        in_num = in_source(distances, rho)
        zero_term = 1e-10
        in0_num, in1_num, in2_num, in3_num = in_num[:-3], in_num[1:-2], in_num[2:-1], in_num[3:]
        d0, d1, d2, d3 = distances[:-3], distances[1:-2], distances[2:-1], distances[3:]
        th0, th1, th2, th3 = jnp.arange(4)
        num_inside  = in1_num * in2_num
        num_in2out  = in1_num * (1.0 - in2_num)
        num_out2in  = (1.0 - in1_num) * in2_num
        th_est      = cubic_interp(rho, d0, d1, d2, d3, th0, th1, th2, th3, epsilon=zero_term)
        frac_in2out = jnp.clip((th_est - th1), 0.0, 1.0)
        frac_out2in = jnp.clip((th2 - th_est), 0.0, 1.0)
        area_inside = r0 * dth * num_inside
        area_crossing = r0 * dth * (num_in2out * frac_in2out + num_out2in * frac_out2in)
        return jnp.sum(area_inside + area_crossing)  

    #@jax.checkpoint
    def _compute_for_range(r_range: Array, th_range: Array) -> Array:
        """Integrate over a given ``(r, theta)`` rectangle using uniform grids.

        Returns the area contribution of the subregion via a rectangle-rule
        sum across the per-radius angular integrals from ``_process_r``.
        """
        r_values = jnp.linspace(r_range[0], r_range[1], r_resolution, endpoint=True)
        th_values = jnp.linspace(th_range[0], th_range[1], th_resolution, endpoint=True)
        #area_r = jax.checkpoint(vmap(lambda r: _process_r(r, th_values, cubic)))(r_values)
        area_r = vmap(lambda r: _process_r(r, th_values))(r_values)
        dr = r_values[1] - r_values[0]
        total_area = dr * jnp.sum(area_r) # trapezoidal integration
        return total_area
    
    #_process_r = jax.checkpoint(_process_r, prevent_cse=True)
    #_compute_for_range = jax.checkpoint(_compute_for_range, prevent_cse=True)
    
    inputs = (r_scan, th_scan)
    if(1): # memory efficient but seems complex implementation for jax.checkpoint.
        def scan_images(carry, inputs):
            r_range, th_range = inputs
            total_area = _compute_for_range(r_range, th_range)
            #total_area = _compute_for_range(r_range, th_range, cubic=cubic)
            return carry + total_area, None
        magnification_unnorm, _ = lax.scan(jax.checkpoint(scan_images), 0.0, inputs, unroll=1)
    if(0): # vmap case. subtle improvement in speed but worse in memory. More careful for chunking size.
        total_areas = vmap(_compute_for_range, in_axes=(0, 0))(r_scan, th_scan)
        magnification_unnorm = jnp.sum(total_areas)
    
    magnification = magnification_unnorm / rho**2 / jnp.pi
    return magnification 

def cubic_interp(
    x: Union[float, Array],
    x0: Union[float, Array],
    x1: Union[float, Array],
    x2: Union[float, Array],
    x3: Union[float, Array],
    y0: Union[float, Array],
    y1: Union[float, Array],
    y2: Union[float, Array],
    y3: Union[float, Array],
    epsilon: float = 1e-12,
) -> Union[float, Array]:
    """Stable 4-point cubic (Lagrange) interpolation with scaling.

    Evaluates the cubic interpolant passing through the four points
    ``(xk, yk)`` for ``k=0..3`` at position ``x``. To improve numerical
    stability when the abscissas are nearly collinear or clustered, the
    abscissa domain is rescaled to ``[0, 1]`` before computing the Lagrange
    basis. Small ``epsilon`` terms guard against division by zero in degenerate
    configurations.

    Parameters
    - x: float/array – Evaluation abscissa.
    - x0, x1, x2, x3: float/array – Sample abscissas.
    - y0, y1, y2, y3: float/array – Sample ordinates corresponding to each
      abscissa.
    - epsilon: float – Small positive value to avoid division by zero in the
      basis denominators.

    Returns
    - y: float/array – Interpolated value at ``x``.

    Notes
    - This function is used to estimate the angular crossing location of the
      source limb within a four-cell angular stencil.
    - For monotonic constraints or fewer samples consider alternative schemes.
    """
    # Implemented algebraically; faster and more memory-efficient than a
    # matrix-based polyfit for JAX transformations.
    x_min = jnp.min(jnp.array([x0, x1, x2, x3]))
    x_max = jnp.max(jnp.array([x0, x1, x2, x3]))
    scale = jnp.maximum(x_max - x_min, epsilon)
    x_hat = (x - x_min) / scale
    x0_hat, x1_hat, x2_hat, x3_hat = (x0 - x_min) / scale, (x1 - x_min) / scale, (x2 - x_min) / scale, (x3 - x_min) / scale
    L0 = ((x_hat - x1_hat) * (x_hat - x2_hat) * (x_hat - x3_hat)) / \
        ((x0_hat - x1_hat + epsilon) * (x0_hat - x2_hat + epsilon) * (x0_hat - x3_hat + epsilon))
    L1 = ((x_hat - x0_hat) * (x_hat - x2_hat) * (x_hat - x3_hat)) / \
        ((x1_hat - x0_hat + epsilon) * (x1_hat - x2_hat + epsilon) * (x1_hat - x3_hat + epsilon))
    L2 = ((x_hat - x0_hat) * (x_hat - x1_hat) * (x_hat - x3_hat)) / \
        ((x2_hat - x0_hat + epsilon) * (x2_hat - x1_hat + epsilon) * (x2_hat - x3_hat + epsilon))
    L3 = ((x_hat - x0_hat) * (x_hat - x1_hat) * (x_hat - x2_hat)) / \
        ((x3_hat - x0_hat + epsilon) * (x3_hat - x1_hat + epsilon) * (x3_hat - x2_hat + epsilon))
    return y0 * L0 + y1 * L1 + y2 * L2 + y3 * L3

if __name__ == "__main__":
    import time
    jax.config.update("jax_enable_x64", True)
    #jax.config.update("jax_debug_nans", True)
    q = 0.05
    s = 1.0
    alpha = jnp.deg2rad(10) 
    tE = 30 
    t0 = 0.0 
    u0 = 0.0 
    rho = 0.06

    nlenses = 2
    a = 0.5 * s
    e1 = q / (1.0 + q)
    _params = {"a": a, "e1": e1}
    x_cm = a * (1.0 - q) / (1.0 + q)

    num_points = 2000
    t  =  jnp.linspace(-1.0*tE, 1.0*tE, num_points)
    tau = (t - t0)/tE
    y1 = -u0*jnp.sin(alpha) + tau*jnp.cos(alpha)
    y2 = u0*jnp.cos(alpha) + tau*jnp.sin(alpha) 
    w_points = jnp.array(y1 + y2 * 1j, dtype=complex)
    test_params = {"q": q, "s": s}  # Lens parameters

    Nlimb = 500
    r_resolution  = 500
    th_resolution = 1000
    cubic = True

    bins_r = 50
    bins_th = 120
    margin_r = 1.0
    margin_th= 1.0

    from microjax.caustics.extended_source import mag_extended_source
    import MulensModel as mm
    def mag_vbbl(w0: complex, rho: float, u1: float = 0.0, accuracy: float = 1e-4) -> Array:
        """Reference magnification via VBBinaryLensing for benchmarking.

        Parameters
        - w0: complex – Source center.
        - rho: float – Source radius.
        - u1: float – Linear limb-darkening coefficient passed to VBBL.
        - accuracy: float – Target accuracy for VBBL integration.

        Returns
        - magnification: float – VBBL magnification for comparison plots.
        """
        a  = 0.5 * s
        e1 = 1.0 / (1.0 + q)
        e2 = 1.0 - e1  
        bl = mm.BinaryLens(e1, e2, 2*a)
        return bl.vbbl_magnification(w0.real, w0.imag, rho, accuracy=accuracy, u_limb_darkening=u1)
    #magn  = lambda w: mag_uniform(w, rho, r_resolution=2000, th_resolution=1000, **test_params, cubic=True)
    @jit
    def mag_mj(w: complex) -> Array:
        """JIT-wrapped microjax uniform finite-source magnification."""
        return mag_uniform(w, rho, s=s, q=q, Nlimb=Nlimb, bins_r=bins_r, bins_th=bins_th,
                           r_resolution=r_resolution, th_resolution=th_resolution, 
                           margin_r = margin_r, margin_th=margin_th, cubic=cubic)
    def chunked_vmap(func: Callable[[complex], Array], data: Array, chunk_size: int) -> Array:
        """Apply ``vmap`` in chunks to reduce peak memory usage."""
        results = []
        for i in range(0, len(data), chunk_size):
            chunk = data[i:i + chunk_size]
            results.append(jax.vmap(func)(chunk))
        return jnp.concatenate(results)

    magn2  = lambda w0: jnp.array([mag_vbbl(w, rho) for w in w0])
    #magn2 =  jit(vmap(magn2, in_axes=(0,)))
    #magn =  jit(vmap(magn, in_axes=(0,)))

    #_ = magn(w_points).block_until_ready()
    @jax.jit
    def scan_mag_mj(w_points: Array) -> Array:
        """Compute magnifications via ``lax.scan`` over input points."""
        def body_fun(carry, w):
            result = mag_mj(w)
            return carry, result
        _, results = lax.scan(body_fun, None, w_points)
        return results

    print("number of data points: %d"%(num_points))
    from microjax.point_source import mag_point_source, critical_and_caustic_curves
    mag_point_source(w_points, s=s, q=q)
    start = time.time()
    mags_poi = mag_point_source(w_points, s=s, q=q)
    mags_poi.block_until_ready()
    end = time.time()
    print("computation time: %.3f sec (%.3f ms per points) for point-source in microjax"%(end-start, 1000*(end - start)/num_points)) 

    from microjax.multipole import _mag_hexadecapole
    z, z_mask = _images_point_source(w_points - x_cm, nlenses=nlenses, **_params)
    _, _ = _mag_hexadecapole(z, z_mask, rho, nlenses=nlenses, **_params) 
    start = time.time()
    z, z_mask = _images_point_source(w_points - x_cm, nlenses=nlenses, **_params)
    mu_multi, delta_mu_multi = _mag_hexadecapole(z, z_mask, rho, nlenses=nlenses, **_params)
    mu_multi.block_until_ready()
    end = time.time()
    print("computation time: %.3f sec (%.3f ms per points) for hexadecapole in microjax"%(end-start, 1000*(end - start)/num_points)) 

    start = time.time()
    magnifications2 = magn2(w_points)
    magnifications2.block_until_ready() 
    end = time.time()
    print("computation time: %.3f sec (%.3f ms per points) for VBBinaryLensing"%(end - start,1000*(end - start)/num_points))

    chunk_size = 500 
    _ = chunked_vmap(mag_mj, w_points, chunk_size).block_until_ready()
    print("start computation with mag_lc_uniform, %d chunk_size, %d rbin, %d thbin"%(chunk_size, r_resolution, th_resolution))
    print("start computation with vmap")
    start = time.time()
    #magnifications = mag_uniform(w_points, rho, s=s, q=q, Nlimb=2000, r_resolution=r_resolution, th_resolution=th_resolution).block_until_ready()
    magnifications = chunked_vmap(mag_mj, w_points, chunk_size).block_until_ready()
    end = time.time()
    print("computation time: %.3f sec (%.3f ms per points) for vmap in microjax"%(end-start, 1000*(end - start)/num_points))
    
    if(0):
        print("start computation with lax.scan")
        start = time.time()
        magnifications = scan_mag_mj(w_points).block_until_ready()
        end = time.time()
        print("computation time: %.3f sec (%.3f ms per points) for lax.scan in microjax"%(end-start, 1000*(end - start)/num_points))
    #print("computation time: %.3f ms per points for lax.scan in microjax" % (1000 * (end - start) / num_points))
    
   
    # Print out the result
    import matplotlib.pyplot as plt
    import matplotlib as mpl
    from mpl_toolkits.axes_grid1.inset_locator import inset_axes
    import matplotlib.ticker as ticker
    import seaborn as sns
    sns.set_theme(style="ticks")

    critical_curves, caustic_curves = critical_and_caustic_curves(nlenses=2, npts=100, s=s, q=q)

    fig, ax_ = plt.subplots(2,1,figsize=(8,6), sharex=True, gridspec_kw=dict(hspace=0.1, height_ratios=[4,1]))
    ax  = ax_[0]
    ax1 = ax_[1]
    ax_in = inset_axes(ax,
        width="60%", height="60%", 
        bbox_transform=ax.transAxes,
        bbox_to_anchor=(0.35, 0.35, 0.6, 0.6)
    )
    ax_in.set_aspect(1)
    ax_in.set(xlabel="$\mathrm{Re}(w)$", ylabel="$\mathrm{Im}(w)$")
    for cc in caustic_curves:
        ax_in.plot(cc.real, cc.imag, color='red', lw=0.7)
    circles = [plt.Circle((xi,yi), radius=rho, fill=False, facecolor=None, ec="blue", zorder=2) 
               for xi, yi in zip(w_points.real, w_points.imag)
               ]
    c = mpl.collections.PatchCollection(circles, match_original=True, alpha=0.5)
    ax_in.add_collection(c)
    ax_in.set_aspect(1)
    ax_in.set(xlim=(-1., 1.2), ylim=(-1.0, 1.))
    ax_in.plot(-q/(1+q) * s, 0 , ".",c="k")
    ax_in.plot((1.0)/(1+q) * s, 0 ,".",c="k")

    ax.plot(t, magnifications, ".", label="microjax", zorder=1)
    ax.plot(t, magnifications2, "-", label="VBBinaryLensing", zorder=2)
    ylim = ax.get_ylim()
    #ax.plot(t, mags_poi, "--", label="point-source", zorder=-1, color="gray")
    ax.set_title("mag_uniform")
    ax.grid(ls=":")
    ax.set_ylabel("magnification")
    ax.plot(t, mags_poi, "--", label="point_source", zorder=-1, color="gray")
    ax.plot(t, mu_multi, ":", label="hexadecapole", zorder=-2, color="orange")
    ax.set_ylim(ylim[0], ylim[1])
    ax1.plot(t, jnp.abs(magnifications - magnifications2)/magnifications2, "-", ms=1)
    #ax1.plot(t, jnp.abs(mags_poi - magnifications2)/magnifications2, "-", ms=1)
    #ax1.plot(t, jnp.abs(mu_multi - magnifications2)/magnifications2, "-", ms=1)
    ax1.grid(ls=":")
    ax1.set_yticks(10**jnp.arange(-4, -2, 1))
    ax1.set_ylabel("relative diff")
    ax1.set_yscale("log")
    ax1.yaxis.set_major_locator(ticker.LogLocator(base=10.0, subs=[1.0, 10**-2, 10**-4, 10**-6], numticks=10))
    ax1.set_ylim(1e-6, 1e-2)
    ax.legend(loc="upper left")
    ax1.set_xlabel("time (days)")
    plt.show()
    #plt.savefig("z_fig/mag_uniform.png", bbox_inches="tight", dpi=300)
    #plt.close()

    if(1):
        diff = jnp.abs(magnifications - magnifications2)/magnifications2
        label = diff > 1e-3
        #for i, (r, th) in enumerate(zip(w_points, diff)):
        for i, (r, th) in enumerate(zip(w_points[label], diff[label])):
            print(i, "%.5f"%(th), r)
        #print("errnous \n", w_points[label], diff[label])



# This file is a modified and extended version of code from the `caustics` package:
#   https://github.com/fbartolic/caustics
# Originally developed by Fran Bartolic under the MIT License.
#
# modifications and extensions have been made by Shota Miyazaki for the `microjax` project.
#
# SPDX-FileCopyrightText: 2022 Fran Bartolic
# SPDX-FileCopyrightText: 2025 Shota Miyazaki
# SPDX-License-Identifier: MIT

__all__ = [
    "match_points",
    "first_nonzero",
    "last_nonzero",
    "first_zero",
    "min_zero_avoiding",
    "max_zero_avoiding",
    "mean_zero_avoiding",
    "sparse_argsort",
]
import jax.numpy as jnp
from jax import lax
import jax

def match_points(a, b):
    """
    Iterate over elements of a in order and find the index of the closest 
    element in b in distance. Return indices which permute b.
    
    This algorithm is not guaranteed to find a permutation of b which minimizes 
    the sum of elementwise distances between every element of a and b 
    (linear sum assignment problem).

    Args:
        a (array_like): 1D array of complex numbers.
        b (array_like): 1D array of complex numbers.

    Returns:
        array_like: Indices which specify a permutation of b.
    """
    # First guess
    vals = jnp.argsort(jnp.abs(b - a[:, None]), axis=1)
    idcs = []
    for i, idx in enumerate(vals[:, 0]):
        # If index is duplicate choose the next best solution
        mask = ~jnp.isin(vals[i], jnp.array(idcs), assume_unique=True)
        idx = vals[i, first_nonzero(mask)]
        idcs.append(idx)

    return jnp.array(idcs)


def first_nonzero(x, axis=0):
    return jnp.argmax(x != 0.0, axis=axis)


def last_nonzero(x, axis=0):
    return lax.cond(
        jnp.any(x, axis=axis),  # if any non-zero
        lambda: (x.shape[axis] - 1)
        - jnp.argmax(jnp.flip(x, axis=axis) != 0, axis=axis),
        lambda: 0,
    )


def first_zero(x, axis=0):
    return jnp.argmax(x == 0.0, axis=axis)


def min_zero_avoiding(x):
    """
    Return the minimum  of a 1D array, avoiding 0.
    """
    x = jnp.sort(x)
    min_x = jnp.min(x)
    cond = min_x == 0.0
    return jnp.where(cond, x[(x != 0).argmax(axis=0)], min_x)


def max_zero_avoiding(x):
    """
    Return the maximum  of a 1D array, avoiding 0.
    """
    x = jnp.sort(x)
    max_x = jnp.max(x)
    cond = max_x == 0.0
    return jnp.where(cond, -min_zero_avoiding(jnp.abs(x)), max_x)


def mean_zero_avoiding(x):
    mask = x == 0.0
    return jnp.where(jnp.all(mask), 0.0, jnp.nanmean(jnp.where(mask, jnp.nan, x)))


def sparse_argsort(a):
    return jnp.where(a != 0, a, jnp.nan).argsort()


def trapz_zero_avoiding(y, x, tail_idx):
    """
    Same as jnp.trapz(y[:tail_idx + 1] x=x[:tail_idx + 1], axis=0).
    """
    I = jax.scipy.integrate.trapezoid(y, x=x)
    #I = jnp.trapz(y, x=x)
    xt, yt = x[tail_idx], y[tail_idx]
    xtp1, ytp1 = x[tail_idx + 1], y[tail_idx + 1]
    return lax.cond(
        tail_idx == len(x) - 1,
        lambda: I,
        lambda: I - 0.5 * ((yt + ytp1) * (xtp1 - xt)),
    )




import numpy as np

from microjax.fastlens.mag_fft_jax import magnification_limb1, magnification_limb2


def test_mag_limb1_A0_matches_A_at_zero_for_large_rho():
    rho = 1e-2
    mag = magnification_limb1()
    a0 = np.array(mag.A0(rho))
    a = np.array(mag.A(0.0, rho))
    assert np.allclose(a, a0, rtol=1e-10, atol=1e-12)


def test_mag_limb2_A0_matches_A_at_zero_for_large_rho():
    rho = 1e-2
    mag = magnification_limb2()
    a0 = np.array(mag.A0(rho))
    a = np.array(mag.A(0.0, rho))
    assert np.allclose(a, a0, rtol=1e-10, atol=1e-12)




# This file is a modified and extended version of code from the `caustics` package:
#   https://github.com/fbartolic/caustics
# Originally developed by Fran Bartolic under the MIT License.
#
# modifications and extensions have been made by Shota Miyazaki for the `microjax` project.
#
# SPDX-FileCopyrightText: 2022 Fran Bartolic
# SPDX-FileCopyrightText: 2025 Shota Miyazaki
# SPDX-License-Identifier: MIT


__all__ = [
    "mag_extended_source",
]

from functools import partial

import numpy as np
import jax.numpy as jnp
from jax import jit, vmap, lax, random

from .integrate import (
    _integrate_unif,
    _integrate_ld,
)
from ..utils import (
    match_points,
    first_nonzero,
    first_zero,
    last_nonzero,
    sparse_argsort,
)

from ..point_source import (
    lens_eq_det_jac,
    _images_point_source,
    _images_point_source_sequential,
)


def _permute_images(z, z_mask, z_parity):
    """
    Sequantially permute the images corresponding to points on the source limb
    starting with the first point such that each point source image is assigned
    to the correct curve. This procedure does not differentiate between real
    and false images, false images are set to zero after the permutation
    operation.
    """
    xs = jnp.stack([z, z_mask, z_parity])

    def apply_match_points(carry, xs):
        z, z_mask, z_parity = xs
        idcs = match_points(carry, z)
        return z[idcs], jnp.stack([z[idcs], z_mask[idcs], z_parity[idcs]])

    init = xs[0, :, 0]
    _, xs = lax.scan(apply_match_points, init, jnp.moveaxis(xs, -1, 0))
    z, z_mask, z_parity = jnp.moveaxis(xs, 1, 0)

    return z.T, z_mask.real.astype(bool).T, z_parity.T


@partial(
    jit,
    static_argnames=(
        "nlenses",
        "npts",
        "niter",
    ),
)
def _images_of_source_limb(
    w0,
    rho,
    nlenses=2,
    npts=300,
    niter=10,
    **params,
):
    key = random.PRNGKey(0)
    key1, key2 = random.split(key)

    def fn(theta, z_init):
        # Add a small perturbation to z_init to avoid situations where
        # there is convergence to the exact same roots when difference in theta
        # is very small and z_init is very close to the exact roots
        u1 = random.uniform(key1, shape=z_init.shape, minval=-1e-6, maxval=1e-6)
        u2 = random.uniform(key2, shape=z_init.shape, minval=-1e-6, maxval=1e-6)
        z_init = z_init + u1 + u2 * 1j

        z, z_mask = _images_point_source(
            rho * jnp.exp(1j * theta) + w0,
            nlenses=nlenses,
            z_init=z_init.T,
            custom_init=True,
            **params,
        )
        det = lens_eq_det_jac(z, nlenses=nlenses, **params)
        z_parity = jnp.sign(det)
        return z, z_mask, z_parity

    # Initial sampling on the source limb
    npts_init = int(0.5 * npts)
    theta = jnp.linspace(-np.pi, np.pi, npts_init - 1, endpoint=False)
    theta = jnp.pad(theta, (0, 1), constant_values=np.pi - 1e-8)
    z, z_mask = _images_point_source_sequential(
        rho * jnp.exp(1j * theta) + w0, nlenses=nlenses, **params
    )
    z_parity = jnp.sign(lens_eq_det_jac(z, nlenses=nlenses, **params))

    # Refine sampling by adding npts_init additional points a fraction
    # 1 / niter at a time
    npts_additional = int(0.5 * npts)
    n = int(npts_additional / niter)

    for i in range(niter):
        # Find the indices (along the contour axis) with the biggest distance
        # gap for consecutive points under the condition that at least
        # one of the two points is a real image
        delta_z = jnp.abs(z[:, 1:] - z[:, :-1])
        delta_z = jnp.where(
            jnp.logical_or(z_mask[:, 1:], z_mask[:, :-1]),
            delta_z,
            jnp.zeros_like(delta_z.real),
        )
        delta_z_max = jnp.max(delta_z, axis=0) # max over images at each phi_n
        idcs_theta = jnp.argsort(delta_z_max)[::-1][:n]

        # Add new points at the midpoints of the top-ranking intervals
        theta_new = 0.5 * (theta[idcs_theta] + theta[idcs_theta + 1])
        z_new, z_mask_new, z_parity_new = fn(theta_new, z[:, idcs_theta])

        # Insert new points
        theta = jnp.insert(theta, idcs_theta + 1, theta_new, axis=0)
        z = jnp.insert(z, idcs_theta + 1, z_new, axis=1)
        z_mask = jnp.insert(z_mask, idcs_theta + 1, z_mask_new, axis=1)
        z_parity = jnp.insert(z_parity, idcs_theta + 1, z_parity_new, axis=1)

    # Get rid of duplicate values of images which may occur in rare cases
    # by adding a very small random perturbation to the images
    z_flat = z.reshape(-1)
    _, ix = jnp.unique(z_flat, return_index=True, size=len(z_flat))
    mask_dup = jnp.full(z_flat.shape, True)
    mask_dup = mask_dup.at[ix].set(False).reshape(z.shape)

    z = jnp.where(
        mask_dup,
        z + random.uniform(key, shape=z.shape, minval=-1e-9, maxval=1e-9),
        z,
    )

    # Permute images
    z, z_mask, z_parity = _permute_images(z, z_mask, z_parity)

    return z, z_mask, z_parity


def _split_segment(segment, n_parts=5):
    """
    Split a single segment such that each part consists of only real images 
    with the same parity such that there are no large jumps in distance between
    consecutive images. `n_parts` specifies the maximum number of output parts.
    """
    z, z_parity, z_mask = segment
    npts = len(z)

    # Construct a mask which evaluates to True at index `i` if the false image
    # mask changes from `i` to `i+1`, if the parity changes or if the distance
    # between the two images is greater than 0.1
    z_diff = z[1:] - z[:-1]
    diff_dist_mask = z_diff.real**2 + z_diff.imag**2 > 0.1**2
    diff_parity = z_parity[1:].real - z_parity[:-1].real
    diff_mask = z_mask[1:].real - z_mask[:-1].real
    changepoints = (diff_dist_mask != False) | (diff_parity != 0) | (diff_mask != False)

    # Expand to sort out the first and the last points
    changepoints = jnp.concatenate([
        jnp.array([jnp.where(z_mask[0] == 0, False, True)]),
        changepoints,
        jnp.array([jnp.where(z_mask[-1] == 0, False, True)]),
    ])
    diff_mask = jnp.concatenate([
        jnp.array([jnp.where(z_mask[0] == 0, 0., 1.)]),
        diff_mask,
        jnp.array([jnp.where(z_mask[-1] == 0, 0., -1.)]),
    ])

    # Find the indices of the starting and ending points of intervals which 
    # contain real images with the same parity
    idcs_start = jnp.where(
        changepoints & (diff_mask >= 0),
        changepoints,
        jnp.zeros_like(changepoints)
    )
    idcs_end = jnp.where(
        changepoints & (diff_mask <= 0),
        changepoints,
        jnp.zeros_like(changepoints)
    )

    idcs_start = jnp.argwhere(idcs_start, size=2*n_parts)[:, 0]
    idcs_end = jnp.argwhere(idcs_end, size=2*n_parts)[:, 0]

    # Split into n_parts parts
    n = jnp.arange(npts)

    def mask_region(carry, xs):
        l, r = xs
        mask = jnp.where(
            (l == 0.0) & (r == 0.0),  # check if region is empty
            jnp.zeros(npts),
            (n >= l) & (n < r),
        )
        return 0, mask

    _, masks = lax.scan(mask_region, 0, (idcs_start, idcs_end))
    segments_split = vmap(lambda mask: segment * mask)(masks)
    return segments_split



def _process_segments(segments, nr_of_segments=20):
    """
    Process raw contour segments such that each segment is contiguous (meaning
    that there are no gaps between the segments head and tail) and that the head
    of the segment is at the 0th index in the array. The resultant number of
    segments is in general greater than the number of images and is at most
    equal to `nr_of_segments`. Segments with fewer than 3 points are removed.
    """
    # Split segments
    segments = jnp.concatenate(vmap(_split_segment)(segments))

    # If a segment consists of just 1 point, set it to zero 
    mask = jnp.sum((segments[:, 0] != 0 + 0j).astype(int), axis=1) < 2
    segments = segments * (~mask[:, None, None])

    # Sort segments such that the nonempty segments appear first and shrink array
    sorted_idcs = jnp.argsort(jnp.any(jnp.abs(segments[:, 0, :]) > 0.0, axis=1))[::-1]
    segments = segments[sorted_idcs]
    segments = segments[:nr_of_segments, :, :]

    # Find head of each segment
    head_idcs = vmap(first_nonzero)(jnp.abs(segments[:, 0, :]))

    # Roll each segment such that head is at idx 0
    segments = vmap(lambda segment, head_idx: jnp.roll(segment, -head_idx, axis=-1))(
        segments, head_idcs
    )

    return segments


def _get_segments(z, z_mask, z_parity, nlenses=2):
    """
    Given the raw images corresponding to a sequence of points on the source
    limb, return two arrays with `open` and `closed` contour segments.
    Closed segments are those images which do not cross the critical curve
    and do not require any extra processing. Open segments need to be stitched 
    together to form closed contours.

    WARNING: The number of these open segments is variable and I assume that
        there are at most 2*n_images open segments. This is a heuristic which
        may potentially fail.

    Args:
        z (array_like): Images corresponding to points on the source limb.
        z_mask (array_like): Mask which indicates which images are real.
        z_parity (array_like): Parity of each image.
        nlenses (int, optional): _description_. Defaults to 2.

    Returns:
        tuple (array_like, array_like, bool): A tuple with the following
        elements:
           array_like: Array with shape (n_images, 2, npts) containing  the
            closed segments.
           array_like: Array with shape (2*n_images, 2, npts) containing  the
            open segments.
           bool: A boolean indicating if the last array is empty (source does
            not cross the critical curve).
    """
    n_images = nlenses**2 + 1
    nr_of_segments = 3 * n_images

    # Apply mask and bundle the images and parity into a single array
    z = z * z_mask
    z_parity = z_parity * z_mask
    segments = jnp.stack([z, z_parity, z_mask])
    segments = jnp.moveaxis(segments, 0, 1)

    # Split segments into segments which are already closed contours (they don't
    # cross the critical curve) and open segments
    mask_closed = (jnp.abs((z[:, 0] - z[:, -1])) < 1e-5) & jnp.all(z_mask, axis=1)

    segments_closed = segments * mask_closed[:, None, None]
    segments_open = segments * (~mask_closed[:, None, None])

    # If there are open segments split them such that each row is a single
    # contigous segment. This expands the size of total number of segments_open
    # to `nr_of_segments`.
    all_closed = jnp.all(mask_closed)
    segments_open = lax.cond(
        all_closed,
        lambda s: jnp.pad(s, ((0, nr_of_segments - s.shape[0]), (0, 0), (0, 0))),
        lambda s: _process_segments(s, nr_of_segments=nr_of_segments),
        segments_open,
    )

    segments_closed = segments_closed[:, :2, :]
    segments_open = segments_open[:, :2, :]

    return segments_closed, segments_open, all_closed


def _concatenate_segments(segment_first, segment_second):
    segment_first_length = first_zero(jnp.abs(segment_first[0]))
    return segment_first + jnp.roll(segment_second, segment_first_length, axis=-1)


def _get_segment_length(segment, tail_idx):
    """Get the physical length of a segment."""
    diff = jnp.diff(segment[0])
    diff = diff.at[tail_idx].set(0.0)
    return jnp.abs(diff).sum()


def _connection_condition(
    seg1, seg2, tidx1, tidx2, ctype, min_dist=1e-05, max_dist=1e-01, max_ang=60.0
):
    """
    Dermine wether two segments should be connected or not for a specific
    type of connection. We differentiate between four types of connections:
       - `ctype` == 0: Tail-Head connection
       - `ctype` == 1: Head-Tail connection
       - `ctype` == 2: Head-Head connection
       - `ctype` == 3: Tail-Tail connection

    We use four criterions to determine if the segments should be connected:
        1. For T-H and H-T the two segments need to have the same parity and
            for H-H and T-T connections they need to have opposite parity.
        2. If we form a line consisting of two points at the end of each segment,
            such that the second point is the connection point, the distance
            between two potential connection points of the segments must be
            less than the distance between the other two points.
        3. The smaller of the two angles formed by the intersection of `line1`
            and `line2` must be less than `max_ang` degrees.
        4. The distance between two potential connection points must be less
            than `max_dist`.

    If the distance between the two connection points is less than `min_dist`,
    and the parity condition is satisfied the function returns `True`
    irrespective of the other conditions.

    All of this is to ensure that we avoid connecting two segments which
    shouldn't be connected.

    Args:
        seg1 (array_like): First segment.
        seg2 (array_like): Second segment.
        tidx1 (int): Index specifying the tail of the first segment.
        tidx2 (int): Index specifying the tail of the second segment.
        ctype (int): Type of connection. 0 = T-H, 1 = H-T, 2 = H-H, 3 = T-T.
        min_dist (float, optional): If the distance between the connection points
            of the two segments is less than this value and the parity
            condition evaluates to `True`, the segments are connected
            irrespective of other conditions.
        max_dist (float, optional): Maximum distance between the ends of the
            two segments. If the distance is greater than this value function
            return `False`. Defaults to 1e-01.
        max_ang (float, optional): Angle in degrees. See criterion 2 above.

    Returns:
        bool: True if the two segments should be connected with a `ctype`
            connection.
    """
    def get_segment_head(seg, t):
        # Sometimes two points at end or beginning are nearly identical, need to
        # avoid those situations by picking the next point
        x = seg[0]
        cond = (jnp.abs(x[1] - x[0]) > 1e-5) | (t <= 1)
        line = lax.cond(
            cond,
            lambda: x[:2],
            lambda: x[1:3],
        )
        return line[::-1]

    def get_segment_tail(seg, t):
        x = seg[0]
        cond = (jnp.abs(x[t] - x[t - 1]) > 1e-5) | (t <= 1)
        line = lax.cond(
            cond,
            lambda: lax.dynamic_slice(x, (t - 1,), (2,)),
            lambda: lax.dynamic_slice(x, (t - 2,), (2,)),
        )
        return line

    # Evaluate parity condition
    same_parity = seg1[1, 0].real * seg2[1, 0].real > 0.0
    conds_parity = jnp.stack(
        [
            same_parity,
            same_parity,
            ~same_parity,
            ~same_parity,
        ]
    )
    cond_parity = conds_parity[ctype]

    # Evaluate conditions 2-4
    line1, line2 = lax.switch(
        ctype,
        [
            lambda s1, s2, t1, t2: (get_segment_tail(s1, t1), get_segment_head(s2, t2)),
            lambda s1, s2, t1, t2: (get_segment_head(s1, t1), get_segment_tail(s2, t2)),
            lambda s1, s2, t1, t2: (get_segment_head(s1, t1), get_segment_head(s2, t2)),
            lambda s1, s2, t1, t2: (
                get_segment_tail(s1, tidx1),
                get_segment_tail(s2, t2),
            ),
        ],
        seg1,
        seg2,
        tidx1,
        tidx2,
    )

    dist = jnp.abs(line1[1] - line2[1])
    cond1 = dist < max_dist

    # Angle between two vectors
    vec1 = (line1[1] - line1[0]) / jnp.abs(line1[1] - line1[0])
    vec2 = (line2[1] - line2[0]) / jnp.abs(line2[1] - line2[0])
    alpha = jnp.arccos(
        jnp.real(vec1) * jnp.real(vec2) + jnp.imag(vec1) * jnp.imag(vec2)
    )
    cond2 = (180.0 - jnp.rad2deg(alpha)) < max_ang

    # Distance between endpoints of line1 and line2 has to be smaller than the
    # distance between points where the line begins
    cond3 = jnp.abs(line1[1] - line2[1]) < jnp.abs(line1[0] - line2[0])

    cond_geom = jnp.logical_or(cond1 & cond2 & cond3, dist < min_dist)

    return cond_parity & cond_geom


def _merge_two_segments(seg1, seg2, tidx1, tidx2, ctype):
    """
    Merge two segments into one assuming that the length of the merged
    segments is equal to at most the shape of `seg1.shape[1]`.

    Args:
        seg1 (array_like): First segment.
        seg2 (array_like): Second segment.
        tidx1 (int): Index specifying the tail of the first segment.
        tidx2 (int): Index specifying the tail of the second segment.
        ctype (int): Type of connection. 0 = T-H, 1 = H-T, 2 = H-H, 3 = T-T.

    Returns
        (array_like, int): The merged segment and the index of the tail of
            the merged segment.
    """

    def hh_connection(seg1, seg2, tidx1, tidx2):
        seg2 = seg2[:, ::-1]  # flip
        seg2 = jnp.roll(seg2, -(seg2.shape[-1] - tidx2 - 1), axis=-1)
        seg2 = seg2.at[1].set(-1 * seg2[1])
        seg_merged = _concatenate_segments(seg2, seg1)
        return seg_merged, tidx1 + tidx2 + 1

    def tt_connection(seg1, seg2, tidx1, tidx2):
        seg2 = seg2[:, ::-1]  # flip
        seg2 = jnp.roll(seg2, -(seg2.shape[-1] - tidx2 - 1), axis=-1)
        seg2 = seg2.at[1].set(-1 * seg2[1])
        seg_merged = _concatenate_segments(seg1, seg2)
        return seg_merged, tidx1 + tidx2 + 1

    def th_connection(seg1, seg2, tidx1, tidx2):
        seg_merged = _concatenate_segments(seg1, seg2)
        return seg_merged, tidx1 + tidx2 + 1

    def ht_connection(seg1, seg2, tidx1, tidx2):
        seg_merged = _concatenate_segments(seg2, seg1)
        return seg_merged, tidx1 + tidx2 + 1

    seg_merged, tidx_merged = lax.switch(
        ctype,
        [th_connection, ht_connection, hh_connection, tt_connection],
        seg1,
        seg2,
        tidx1,
        tidx2,
    )
    return seg_merged, tidx_merged


def _merge_open_segments(
    segments,
    max_nr_of_contours=3,
    max_nr_of_segments_in_contour=10,
):
    """
    Sequentially merge open segments until they form a closed contour. The
    merging algorithm is as follows:

        1. Select shortest (in length) open segment, set it as the active segment.
        2. Find segments closest to current segment in (H-H, T-T, T-H and H-T),
            select 4 of the closest segments.
        3. Evaluate the `_connection_condition` for the active segment and each
            of the 4 closest segments. Pick the segment for which the condition
            is True and the distance is minimized.
        4. Merge the active segment with the selected segment.
        5. Repeat steps 2-4 until the active segment becomes a closed contour.
            This will happen when there are no more open segments which satisfy
            the connection condition.

    The whole process is repeated `max_nr_of_contours` times.

    Args:
        segments (array_like): Array containing the segments, shape
            (`n_segments`, 2, `n_points`).
        segments_mask (array_like): Mask indicating which segments are closed to
            begin with. Shape (`n_segments`).
        max_nr_of_contours (int): Maximum number of contours comprised of the
            input segments.
        max_nr_of_segments_in_contour (int): Maximum number of segments for one
            contour. Should be at least 12 in case of triple lensing. Default is
            10.
        max_dist (float): Maximum distance between two segment end points that
            are allowed to be connected.

    Returns:
        array_like: Array with shape (`n_segments`, 2, `n_points`) containing
        the merged segments contours.
    """

    def merge_with_another_segment(seg_active, tidx_active, segments, tidcs):
        # Compute all T-H, H-T, H-H, T-T distances between the current segment
        # and all other open segments
        dist_th = jnp.abs(seg_active[0, tidx_active] - segments[:, 0, 0])
        dist_ht = vmap(lambda seg, tidx: jnp.abs(seg_active[0, 0] - seg[0, tidx]))(
            segments, tidcs
        )
        dist_hh = jnp.abs(seg_active[0, 0] - segments[:, 0, 0])
        dist_tt = vmap(
            lambda seg, tidx: jnp.abs(seg_active[0, tidx_active] - seg[0, tidx])
        )(segments, tidcs)

        distances = jnp.stack([dist_th, dist_ht, dist_hh, dist_tt])

        # Get the connection type and the index of the "closest" segment, the
        # second closest segment and so on
        ctype1, idx1 = jnp.unravel_index(
            jnp.argsort(distances.reshape(-1))[0], distances.shape
        )
        ctype2, idx2 = jnp.unravel_index(
            jnp.argsort(distances.reshape(-1))[1], distances.shape
        )
        ctype3, idx3 = jnp.unravel_index(
            jnp.argsort(distances.reshape(-1))[2], distances.shape
        )
        ctype4, idx4 = jnp.unravel_index(
            jnp.argsort(distances.reshape(-1))[3], distances.shape
        )

        # Evaluate the connection conditions for each of the four closest segments
        success1 = _connection_condition(
            seg_active, segments[idx1], tidx_active, tidcs[idx1], ctype1
        )
        success2 = _connection_condition(
            seg_active, segments[idx2], tidx_active, tidcs[idx2], ctype2
        )
        success3 = _connection_condition(
            seg_active, segments[idx3], tidx_active, tidcs[idx3], ctype3
        )
        success4 = _connection_condition(
            seg_active, segments[idx4], tidx_active, tidcs[idx4], ctype4
        )

        def branch1(segments, tidcs):
            # Select the closest segment that satisfies the connection condition
            idx_best = first_nonzero(
                jnp.array([success1, success2, success3, success4]).astype(float)
            )
            seg_best = jnp.stack(
                [segments[idx1], segments[idx2], segments[idx3], segments[idx4]]
            )[idx_best]
            tidx_best = jnp.stack([tidcs[idx1], tidcs[idx2], tidcs[idx3], tidcs[idx4]])[
                idx_best
            ]
            ctype = jnp.stack([ctype1, ctype2, ctype3, ctype4])[idx_best]

            # Merge that segment with the active segment
            seg_active_new, tidx_active_new = _merge_two_segments(
                seg_active,
                seg_best,
                tidx_active,
                tidx_best,
                ctype,
            )

            # Zero-out the segment that was merged
            idx_seg = jnp.array([idx1, idx2, idx3, idx4])[idx_best]
            segments = segments.at[idx_seg].set(
                jnp.array(jnp.zeros_like(segments[0]), segments[idx1])
                #jnp.zeros_like(segments[0]), segments[idx1]
            )
            return seg_active_new, tidx_active_new, segments, tidcs

        def branch2(segments, tidcs):
            return seg_active, tidx_active, segments, tidcs

        return lax.cond(
            jnp.any(jnp.array([success1, success2, success3, success4])),
            branch1,
            branch2,
            segments,tidcs
        )


    def body_fn(carry, _):
        # Get the active segment, the index of its tail and all the other
        # open segments and their tail indices from previous iteration
        seg_active, tidx_active, segments, tidcs = carry

        # If all segments are empty don't do anything
        stopping_criterion = ~jnp.any(segments[:, 0, 0])
        seg_active, tidx_active, segments, tidcs = lax.cond(
            stopping_criterion,
            lambda: (seg_active, tidx_active, segments, tidcs),
            lambda: merge_with_another_segment(
                seg_active, tidx_active, segments, tidcs
            ),
        )

        return (
            seg_active,
            tidx_active,
            segments,
            tidcs,
        ), 0.0

    # Compute tail index for each segment
    tail_idcs = vmap(last_nonzero)(segments[:, 0, :].real)

    # Pad segments with zeros to make room for merged segments as this array
    # be modified in place
    # WARNING: this will fail silently if a contour is longer than 3*segments.shape[-1]
    segments = jnp.pad(
        segments, ((0, 0), (0, 0), (0, 3 * segments.shape[-1])), constant_values=0.0
    )

    # Merge all open segments: start by selecting 0th open segment and then
    # sequentially merging it with other ones until a closed segment is formed
    segments_merged_list = []

    for i in range(max_nr_of_contours):
        # Sort open segments such that the shortest segments appear first
        segment_lengths = vmap(_get_segment_length)(segments, tail_idcs)
        _idcs = sparse_argsort(segment_lengths)
        segments, tail_idcs = segments[_idcs], tail_idcs[_idcs]

        idcs = jnp.arange(0, max_nr_of_segments_in_contour)
        init = (segments[0], tail_idcs[0], segments[1:], tail_idcs[1:])
        carry, _ = lax.scan(body_fn, init, idcs)
        seg_merged, tidx_merged, segments, tail_idcs = carry
        segments_merged_list.append(seg_merged)
        max_nr_of_segments_in_contour -= 2

    return jnp.stack(segments_merged_list)


def _contours_from_closed_segments(segments):
    """
    Process closed segments by extracting the parity information and adding a
    single point at the end so that the last point is the same as the first.
    """
    # Extract per contour parity values
    contours_p = segments[:, 1, 0].real

    # Remove parity dimension
    contours = segments[:, 0]

    # Set the last point in each contour to be equal to the first point
    contours = jnp.hstack([contours, contours[:, 0][:, None]])

    return contours, contours_p


def _contours_from_open_segments(
    segments,
    max_nr_of_contours=3,
    max_nr_of_segments_in_contour=20,
):
    """
    Given a set of open image segments, some of which may be open, return
    closed contours which are obtained by merging the open segments together.

    Args:
        segments (array_like): Array of shape `(n_segments, 2, n_points)`
            containing the segment points and the parity for each point.
        n_contours (int, optional): Final number of contours. Defaults to 5.
        max_nr_of_segments_in_contour (int): Maximum number of segments for a
            closed contour. Should be at least 12 in case of triple lensing.
            Default is 10.

    Returns:
        tuple: A tuple of (contours, parity) where `contours` is an array with
        shape `(n_contours, n_points)` and parity is an array of shape `n_contours`
        indicating the parity of each contour.
    """

    # Run the merging algorithm
    segments_merged = _merge_open_segments(
        segments,
        max_nr_of_contours=max_nr_of_contours,
        max_nr_of_segments_in_contour=max_nr_of_segments_in_contour,
    )

    # Extract per contour parity values
    contours_p = segments_merged[:, 1, 0].real

    # Remove the parity dimension
    contours = segments_merged[:, 0]

    # Set the last point in each contour to be equal to the first point
    tail_idcs = vmap(last_nonzero)(contours.real)
    contours = vmap(lambda idx, c: c.at[idx + 1].set(c[0]))(tail_idcs, contours)

    return contours, contours_p


@partial(
    jit,
    static_argnames=(
        "nlenses",
        "npts_limb",
        "limb_darkening",
        "npts_ld",
    ),
)
def mag_extended_source(
    w0,
    rho,
    nlenses=2,
    npts_limb=150,
    limb_darkening=False,
    u1=0.0,
    npts_ld=100,
    **params,
):
    """
    Compute the magnification of an extended source with radius `rho` for a 
    system with `nlenses` lenses. If `nlenses` is 2 (binary lens) or 3 
    (triple lens), the coordinate system is set up such that the the origin is 
    at the center of mass of the first two lenses which are both located on the 
    real line. The location of the first lens is $-sq/(1 + q)$ and the second lens 
    is at $s/(1 + q)$. The optional third lens is located at an arbitrary position 
    in the complex plane $r_3e^{-i*\psi}$. The magnification is computed using 
    contour integration in the image plane. Boolean flag `limb_darkening` 
    indicated whether linear limb-darkening needs to taken into account. If 
    `limb_darkening` is set to True the linear limb-darkening coefficient 
    u1 needs to be specified as well. 

    If `nlenses` is 2 only the parameters `s` and `q` should be specified. If 
    `nlenses` is 3, the parameters `s`, `q`, `q3`, `r3` and `psi` should be 
    specified.

    !!! note

        Turning on limb-darkening (`limb_darkening=True`) slows down the 
        computation by up to an order of magnitude.


    Args:
        w0 (complex): Source position in the complex plane.
        rho (float): Source radius in Einstein radii.
        nlenses (int): Number of lenses in the system.
        npts_limb (int, optional): Initial number of points uniformly distributed
            on the source limb when computing the point source magnification.
            The final number of points is greater than this value because
            the number of points is decreased geometrically by a factor of
            1/2 until it reaches 2.
        limb_darkening (bool, optional): If True, compute the magnification of
            a limb-darkened source. If limb_darkening is enabled the u1 linear
            limb-darkening coefficient needs to be specified. Defaults to False.
        u1 (float, optional): Linear limb darkening coefficient. Defaults to 0..
        npts_ld (int, optional): Number of points at which the stellar brightness
            function is evaluated when computing contour integrals 
            $\int P(z_1^\prime, z_2) dz_1^\prime$ and 
            $\int Q(z_1, z_2^\prime) dz_2^\prime$ (see Dominik 1998). Defaults 
            to 100.
        s (float): Separation between the two lenses. The first lens is located 
            at $-sq/(1 + q)$ and the second lens is at $s/(1 + q)$ on the real line.
        q (float): Mass ratio defined as $m_2/m_1$.
        q3 (float): Mass ratio defined as $m_3/m_1$.
        r3 (float): Magnitude of the complex position of the third lens.
        psi (float): Phase angle of the complex position of the third lens.
        roots_itmax (int, optional): Number of iterations for the root solver.
        roots_compensated (bool, optional): Whether to use the compensated
            arithmetic version of the Ehrlich-Aberth root solver.

    Returns:
        float: Total magnification at the source position `w0`.
    """
    if nlenses == 1:
        _params = {}
    elif nlenses == 2:
        s, q = params["s"], params["q"]
        a = 0.5*s
        e1 = q/(1 + q)
        _params = {"a": a, "e1": e1}

        # Shift w by x_cm
        x_cm = a*(1 - q)/(1 + q)
        w0 -= x_cm #miyazaki
        #w0 += x_cm
    elif nlenses == 3:
        s, q, q3, r3, psi = params["s"], params["q"], params["q3"], params["r3"], params["psi"]
        a = 0.5*s
        e1 = q/(1 + q + q3)         #miyazaki
        e2 = (1-q3)/(1 + q + q3)    #miyazaki
        #e1 = q/(1 + q + q3)
        #e2 = q*e1
        r3 = r3*jnp.exp(1j*psi)
        _params = {"a": a, "r3": r3, "e1": e1, "e2": e2}

        # Shift w by x_cm
        x_cm = a*(1 - q)/(1 + q)
        w0 -= x_cm #miyazaki
        #w0 += x_cm
    else:
        raise ValueError("`nlenses` has to be set to be <= 3.")

    # Get ordered point source images at the source limb
    z, z_mask, z_parity = _images_of_source_limb(
        w0,
        rho,
        nlenses=nlenses,
        npts=npts_limb,
        **_params,
    )

    # Integration function depending on whether the source is limb-darkened
    if limb_darkening:
        integrate = lambda contour, tidx: _integrate_ld(
            contour,
            tidx,
            w0,
            rho,
            u1=u1,
            nlenses=nlenses,
            npts=npts_ld,
            **_params,
        )

    else:
        integrate = lambda contour, tidx: _integrate_unif(contour, tidx)

    # For N = 1 the contours are trivially to obtain
    if nlenses == 1:
        contours, contours_p = _contours_from_closed_segments(
            jnp.moveaxis(jnp.stack([z, z_parity]), 0, 1)
        )
        tail_idcs = jnp.array([z.shape[1] - 1, z.shape[1] - 1])
        I = vmap(integrate)(contours, tail_idcs)
        return jnp.abs(jnp.sum(I * contours_p)) / (np.pi * rho**2)

    # For N = 2 we first have to obtain segments and then convert those to
    # closed contours
    elif (nlenses == 2) or (nlenses == 3):
        max_nr_of_contours = 3
        # Get segments. If `all_closed` is True there are no caustic crossings
        # and everything is easy
        segments_closed, segments_open, all_closed = _get_segments(
            z, z_mask, z_parity, nlenses=nlenses
        )

        # Get contours from closed segments
        contours1, contours_p1 = _contours_from_closed_segments(segments_closed)

        # Integrate over contours obtained from closed segments
        tail_idcs = jnp.repeat(contours1.shape[1] - 1, contours1.shape[0])
        I1 = vmap(integrate)(contours1, tail_idcs)
        mags1 = I1 * contours_p1

        # If there are caustic crossings things are a lot more complicated and
        # we have to stitch together the open segments to form closed contours.
        branch1 = lambda _: jnp.zeros(max_nr_of_contours)

        def branch2(segments):
            contours, contours_p = _contours_from_open_segments(
                segments, max_nr_of_contours=max_nr_of_contours
            )
            tail_idcs = vmap(last_nonzero)(contours.real)
            I = vmap(integrate)(contours, tail_idcs)
            return I * contours_p

        mags2 = lax.cond(all_closed, branch1, branch2, segments_open)
        mag = jnp.abs(mags1.sum() + mags2.sum()) / (np.pi * rho**2)

        return mag

    else:
        raise ValueError("`nlenses` has to be set to be <= 3.")



from scipy.special import ellipe as ellipe_s
from scipy.special import ellipk as ellipk_s
from microjax.fastlens.special import ellipe
from microjax.fastlens.special import ellipk
import numpy as np
import jax.numpy as jnp
from jax import jit, vmap

from jax import config

config.update("jax_enable_x64", True)


def test_ellipe():
    """This function tests the ellipe function by comparing the results of the ellipe function from microjax and scipy."""
    random_m_values = np.random.uniform(low=-1, high=1, size=1000)
    random_m_values_jax = jnp.array(random_m_values)
    ellipe_vmap = jit(vmap(ellipe))
    ellipe_jax = ellipe_vmap(random_m_values_jax)
    ellipe_scipy = ellipe_s(random_m_values_jax)
    atol = 1e-14

    assert jnp.allclose(ellipe_jax, ellipe_scipy, atol=atol)


def test_ellipk():
    """This function tests the ellipk function by comparing the results of the ellipk function from microjax and scipy."""
    random_m_values = np.random.uniform(low=-1, high=1, size=1000)
    random_m_values_jax = jnp.array(random_m_values)
    ellipk_vmap = jit(vmap(ellipk))
    ellipk_jax = ellipk_vmap(random_m_values_jax)
    ellipk_scipy = ellipk_s(random_m_values_jax)
    atol = 1e-14

    assert jnp.allclose(ellipk_jax, ellipk_scipy, atol=atol)


if __name__ == "__main__":
    test_ellipe()
    test_ellipk()
    print("All tests passed!")



import os

# Force JAX to use CPU to avoid GPU plugin initialization in CI
os.environ.setdefault("JAX_PLATFORMS", "cpu")

try:
    from jax import config
    config.update("jax_enable_x64", True)
except Exception:
    pass




import numpy as np
import jax.numpy as jnp
from jax import config

config.update("jax_enable_x64", True)

from microjax.point_source import (
    mag_point_source,
    _images_point_source,
    critical_and_caustic_curves,
    lens_eq,
)


def analytic_A_single(u):
    u = np.abs(u)
    return (u**2 + 2.0) / (u * np.sqrt(u**2 + 4.0))


def test_single_lens_magnification_basic_properties():
    # Basic properties: monotone decreasing in |w| and tends to 1 as |w|->inf
    ws = jnp.array([0.1 + 0j, 1.0 + 0j, 3.0 + 0j, 100.0 + 0j])
    A_num = np.array(mag_point_source(ws, nlenses=1))
    # monotone decreasing
    assert np.all(np.diff(A_num) < 0)
    # asymptotic to 1
    assert np.isclose(A_num[-1], 1.0, rtol=5e-4, atol=5e-4)


def test_single_lens_shape_scalar_and_vector():
    w_scalar = 0.5 + 0j
    w_vec = jnp.array([0.5 + 0j, 1.5 + 0j])
    A_s = mag_point_source(w_scalar, nlenses=1)
    A_v = mag_point_source(w_vec, nlenses=1)
    assert np.array(A_s).ndim == 0
    assert np.array(A_v).shape == (2,)


def test_images_satisfy_lens_eq_single():
    w = 0.7 + 0.3j
    z, mask = _images_point_source(w, nlenses=1)
    # Verify masked images satisfy lens eq: y = x - 1/conj(x)
    residual = np.array(lens_eq(np.array(z), nlenses=1) - np.array(w))
    ok = np.array(mask)
    assert residual[ok].size >= 1
    assert np.all(np.abs(residual[ok]) < 1e-9)


def test_binary_far_field_magnification_approx_one():
    # Far from lenses magnification -> 1
    params = {"s": 1.2, "q": 0.3}
    w = jnp.array([50.0 + 0j, 100.0 + 0j])
    A = np.array(mag_point_source(w, nlenses=2, **params))
    assert np.allclose(A, 1.0, rtol=1e-6, atol=1e-6)


def test_critical_caustic_curves_single_shapes():
    n = 32
    z_cr, z_ca = critical_and_caustic_curves(npts=n, nlenses=1)
    z_cr = np.array(z_cr)
    z_ca = np.array(z_ca)
    assert z_cr.shape == (n,)
    assert z_ca.shape == (n,)
    # Unit circle for critical curve; caustic at origin
    assert np.allclose(np.abs(z_cr), 1.0, rtol=1e-12, atol=1e-12)
    assert np.allclose(z_ca, 0.0 + 0.0j, atol=1e-12)



import numpy as np
import jax.numpy as jnp
from jax import config

config.update("jax_enable_x64", True)

from microjax.fastlens.special import j0, j1, j2, j1p5, ellipk, ellipe
from scipy.special import j0 as j0_s, j1 as j1_s, jn as jn_s, jv as jv_s
from scipy.special import ellipk as ellipk_s, ellipe as ellipe_s


def test_bessel_at_zero_and_small():
    # Small arguments agree with SciPy; include zero
    xs = jnp.array([0.0, 1e-6, 1e-4, 1e-2])
    # Known limits at zero where defined
    assert np.isclose(np.array(j0(xs))[0], 1.0, atol=1e-15)
    assert np.isclose(np.array(j1(xs))[0], 0.0, atol=1e-15)
    assert jnp.allclose(j0(xs), j0_s(xs), atol=1e-12)
    assert jnp.allclose(j1(xs), j1_s(xs), atol=1e-12)
    xs_nz = jnp.array([1e-12, 1e-8, 1e-4, 1e-2])
    assert jnp.allclose(j2(xs_nz), jn_s(2, xs_nz), atol=1e-12)
    assert float(j2(xs_nz)[0]) < 1e-9
    assert jnp.allclose(j1p5(xs_nz), jv_s(1.5, xs_nz), atol=1e-12)


def test_elliptic_edges():
    # m=0: both are pi/2 (use scalar to avoid carry shape mismatch path)
    m0 = 0.0
    assert np.isclose(np.array(ellipk(m0)), ellipk_s(m0), atol=1e-14)
    assert np.isclose(np.array(ellipe(m0)), ellipe_s(m0), atol=1e-14)

    # small +/- m should match SciPy closely; iterate scalars
    for m in [-1e-6, -1e-4, 1e-6, 1e-4]:
        assert np.isclose(np.array(ellipk(m)), ellipk_s(m), rtol=1e-10, atol=1e-12)
        assert np.isclose(np.array(ellipe(m)), ellipe_s(m), rtol=1e-10, atol=1e-12)



import numpy as np

from microjax.fastlens.mag_fft import magnification_disk, A_point


def test_magnification_disk_A0_large_rho_matches():
    # Use rho larger than rho_switch to hit the large-rho FFT path
    rho = 1e-2
    mag = magnification_disk()
    a0_expected = np.sqrt(rho**2 + 4) / rho
    # At u=0, large-rho branch returns exactly A0(rho)
    a = mag.A(0.0, rho)
    assert np.allclose(a, a0_expected, rtol=1e-10, atol=1e-12)


def test_magnification_disk_small_rho_limits():
    # Small rho uses approximation: point-source for u >= u_switch * rho
    rho = 1e-6
    mag = magnification_disk()
    u = 100 * rho  # greater than u_switch * rho (default u_switch=10)
    a = mag.A(u, rho)
    assert np.allclose(a, A_point(u), rtol=1e-6, atol=1e-8)


def test_A_point_asymptotics():
    # Sanity: A_point(u) -> 1 + 1/(2 u^4) as u -> inf; test near 1
    u = np.array([10.0, 30.0])
    ap = A_point(u)
    assert np.all(ap > 1.0)
    # Closer to 1 for larger u
    assert (ap[1] - 1.0) < (ap[0] - 1.0)




import numpy as _np
import pytest
from microjax.poly_solver import poly_roots


def _sort_by_angle(z, jnp):
    # Stable sort for complex roots: by angle then magnitude
    ang = jnp.angle(z)
    mag = jnp.abs(z)
    order = jnp.lexsort((mag, ang))
    return z[order]


def test_poly_roots_quadratic():
    import os
    os.environ["JAX_PLATFORMS"] = "cpu"
    import jax
    from jax import config
    import jax.numpy as jnp
    config.update("jax_enable_x64", True)

    # (z-1)(z+2) = z^2 + z - 2 -> coeffs [1, 1, -2]
    try:
        coeffs = jnp.array(_np.array([1.0 + 0j, 1.0 + 0j, -2.0 + 0j]))
        roots = _sort_by_angle(poly_roots(coeffs[None, :])[0], jnp)
        expected = _sort_by_angle(jnp.array(_np.array([1.0 + 0j, -2.0 + 0j])), jnp)
        assert jnp.allclose(roots, expected, atol=1e-9)
    except RuntimeError as e:
        if "Unable to initialize backend" in str(e):
            pytest.skip("Skipping on non-CPU JAX backend environment")
        raise


def test_poly_roots_cubic():
    import os
    os.environ["JAX_PLATFORMS"] = "cpu"
    import jax
    from jax import config
    import jax.numpy as jnp
    config.update("jax_enable_x64", True)

    # (z-1)(z-2)(z-3) = z^3 - 6z^2 + 11z - 6
    try:
        coeffs = jnp.array(_np.array([1.0 + 0j, -6.0 + 0j, 11.0 + 0j, -6.0 + 0j]))
        roots = _sort_by_angle(poly_roots(coeffs[None, :])[0], jnp)
        expected = _sort_by_angle(jnp.array(_np.array([1.0 + 0j, 2.0 + 0j, 3.0 + 0j])), jnp)
        assert jnp.allclose(roots, expected, atol=1e-8)
    except RuntimeError as e:
        if "Unable to initialize backend" in str(e):
            pytest.skip("Skipping on non-CPU JAX backend environment")
        raise



import jax.numpy as jnp
from jax import jit, lax, vmap
from functools import partial
from jax import custom_jvp

# from jax.scipy.special import digamma
# from jax.lax import digamma
# from microjax.fastlens.gamma_jax import gamma_jax as gamma

__all__ = ["gamma", "ellipk", "ellipe", "j0", "j1", "j2", "j1p5"]


@custom_jvp
def gamma_(z):
    """
    Custom gamma function with JAX similar to scipy.special.gamma.
    I see that the accuracy compared with scipy.special.gamma is met at 1e-10 within -100<Re(z)<100 and -100<Im(z)<100.
    I adopted the algorothm from https://ui.adsabs.harvard.edu/abs/2021arXiv210400697C/abstract.

    Arg   : z (complex)
    return: Gamma(z) (complex)
    """
    K = 5
    N = 16
    c = jnp.log(2 * jnp.pi) / 2
    a = jnp.array([1 / 12, -1 / 360, 1 / 1260, -1 / 1680, 1 / 1188])

    z = jnp.atleast_1d(z)
    original_shape = z.shape
    z = z.ravel()

    negative_real_mask = jnp.real(z) <= 0
    not_special_case = ~negative_real_mask

    def compute_case(z_val):
        t = z_val + N
        gam = a[0] / t
        t_squared = t**2
        for k in range(1, K):
            t *= t_squared
            gam += a[k] / t
        u = jnp.prod(z_val[:, None] + jnp.arange(N), axis=1)
        lg = c - (z_val + N) + (z_val + N - 0.5) * jnp.log(z_val + N) - jnp.log(u) + gam
        return jnp.exp(lg)

    def compute_reflection(z_val):
        reflected_z = 1 - z_val
        sin_term = jnp.sin(jnp.pi * z_val)
        gamma_reflected = compute_case(reflected_z)
        return jnp.pi / (sin_term * gamma_reflected)

    # Compute results for all elements
    positive_results = compute_case(z)
    negative_results = compute_reflection(z)

    # Select the appropriate results based on the masks
    g = jnp.where(not_special_case, positive_results, negative_results)

    non_positive_integer_mask = negative_real_mask & (
        jnp.real(z) == jnp.floor(jnp.real(z))
    )
    g = jnp.where(non_positive_integer_mask, jnp.nan, g)

    return g.reshape(original_shape)


def digamma_(z):
    """
    Custom digamma function with JAX, implemented from the Julia algorithm
    Args:
        z (complex): Input value(s).
    Returns:
        complex: Digamma of z.
    """
    # constants
    coeffs = jnp.array(
        [
            0.08333333333333333,
            -0.008333333333333333,
            0.003968253968253968,
            -0.004166666666666667,
            0.007575757575757576,
            -0.021092796092796094,
            0.08333333333333333,
            -0.4432598039215686,
        ],
        dtype=jnp.float64,
    )
    z = jnp.atleast_1d(z)
    original_shape = z.shape
    z = z.ravel()

    # Reflection for negative real part
    negative_real_mask = jnp.real(z) < 0
    reflection = jnp.pi / jnp.tan(jnp.pi * z)
    z = jnp.where(negative_real_mask, 1.0 - z, z)
    psi = jnp.where(negative_real_mask, -reflection, 0.0 + 0.0j)

    # Adjustment for x < X
    X = 8.0
    x = jnp.real(z)
    small_mask = x < X
    n_add = jnp.sum(1.0 / (z[:, None] + jnp.arange(X)), axis=-1)
    z = jnp.where(small_mask, z + X, z)
    psi = jnp.where(small_mask, psi - n_add, psi)

    # Asymptotic expansion
    t = 1 / z
    psi += jnp.log(z) - 0.5 * t
    t *= t  # 1/z^2
    psi -= t * jnp.polyval(coeffs[::-1], t)

    return psi.reshape(original_shape)


@gamma_.defjvp
def gamma_jvp(primals, tangents):
    (x,) = primals
    (x_dot,) = tangents
    primal_out = gamma_(x)
    tangent_out = primal_out * digamma_(x) * x_dot
    return primal_out, tangent_out


gamma = jit(gamma_)


def agm(a, b, num_iter=5):
    def agm_step(carry, _):
        a, b = carry
        return ((a + b) / 2, jnp.sqrt(a * b)), None

    (a, b), _ = lax.scan(agm_step, (a, b), None, length=num_iter)
    return a


@custom_jvp
def ellipk_(m):
    num_iter = 5
    a, b = 1.0, jnp.sqrt(1 - m)
    c = agm(a, b, num_iter=num_iter)
    return jnp.pi / (2 * c)


@ellipk_.defjvp
def ellipk_jvp(primals, tangents):
    (m,) = primals
    (m_dot,) = tangents
    k_val = ellipk_(
        m,
    )
    # Derivative of the complete elliptic integral of the first kind with respect to m
    dk_dm = (ellipe_(m) - (1 - m) * k_val) / (2 * m * (1 - m))
    return k_val, m_dot * dk_dm


def agm2(a0, b0, s_sum0, num_iter=5):
    def agm_step2(carry, _):
        a, b, s_sum, n = carry
        a_next, b_next = (a + b) / 2, jnp.sqrt(a * b)
        c_next = 0.5 * (a - b)
        s_sum_next = s_sum + 2 ** (n - 1) * c_next**2
        n += 1
        return (a_next, b_next, s_sum_next, n), None

    n = 1
    (a, b, s_sum, n), _ = lax.scan(
        agm_step2, (a0, b0, s_sum0, n), None, length=num_iter
    )
    return a, s_sum


@custom_jvp
def ellipe_(m):
    num_iter = 5
    a0, b0 = 1.0, jnp.sqrt(1 - m)
    c0 = jnp.sqrt(jnp.abs(a0**2 - b0**2))
    # c0     = jnp.sqrt(a0**2 - b0**2)
    s_sum0 = 0.5 * c0**2
    a, s_sum = agm2(a0, b0, s_sum0, num_iter=num_iter)
    return jnp.pi / (2 * a) * (1 - s_sum)


@ellipe_.defjvp
def ellipe_jvp(primals, tangents):
    (m,) = primals
    (m_dot,) = tangents
    e_val = ellipe_(m)
    k_val = ellipk_(m)
    # Derivative of the complete elliptic integral of the second kind with respect to m
    de_dm = (e_val - k_val) / (2 * m)
    return e_val, m_dot * de_dm


# @jit
"""
def ellipk(m):
    return lax.cond(m > 0, 
                    lambda _: ellipk_(m), 
                    lambda _: 1.0 / jnp.sqrt(jnp.abs(m) + 1.0) * ellipk_(jnp.abs(m) / (jnp.abs(m) + 1.0)), 
                    None)
#@jit
def ellipe(m):
    return lax.cond(m > 0, 
                    lambda _: ellipe_(m), 
                    lambda _: jnp.sqrt(jnp.abs(m) + 1.0) * ellipe_(jnp.abs(m) / (jnp.abs(m) + 1.0)), 
                    None)
"""


def ellipk(m):
    result = jnp.where(
        m > 0,
        ellipk_(m),
        1.0 / jnp.sqrt(jnp.abs(m) + 1.0) * ellipk_(jnp.abs(m) / (jnp.abs(m) + 1.0)),
    )
    return result


def ellipe(m):
    result = jnp.where(
        m > 0,
        ellipe_(m),
        jnp.sqrt(jnp.abs(m) + 1.0) * ellipe_(jnp.abs(m) / (jnp.abs(m) + 1.0)),
    )
    return result


# ellipk = jit(vmap(ellipk, in_axes=(0,)))
# ellipe = jit(vmap(ellipe, in_axes=(0,)))


# I imported the code from https://github.com/google/jax/pull/17038/files
# polynomial coefficients for J1

RP1 = [
    -8.99971225705559398224e8,
    4.52228297998194034323e11,
    -7.27494245221818276015e13,
    3.68295732863852883286e15,
]
RQ1 = [
    1.0,
    6.20836478118054335476e2,
    2.56987256757748830383e5,
    8.35146791431949253037e7,
    2.21511595479792499675e10,
    4.74914122079991414898e12,
    7.84369607876235854894e14,
    8.95222336184627338078e16,
    5.32278620332680085395e18,
]

PP1 = [
    7.62125616208173112003e-4,
    7.31397056940917570436e-2,
    1.12719608129684925192e0,
    5.11207951146807644818e0,
    8.42404590141772420927e0,
    5.21451598682361504063e0,
    1.00000000000000000254e0,
]
PQ1 = [
    5.71323128072548699714e-4,
    6.88455908754495404082e-2,
    1.10514232634061696926e0,
    5.07386386128601488557e0,
    8.39985554327604159757e0,
    5.20982848682361821619e0,
    9.99999999999999997461e-1,
]

QP1 = [
    5.10862594750176621635e-2,
    4.98213872951233449420e0,
    7.58238284132545283818e1,
    3.66779609360150777800e2,
    7.10856304998926107277e2,
    5.97489612400613639965e2,
    2.11688757100572135698e2,
    2.52070205858023719784e1,
]
QQ1 = [
    1.0,
    7.42373277035675149943e1,
    1.05644886038262816351e3,
    4.98641058337653607651e3,
    9.56231892404756170795e3,
    7.99704160447350683650e3,
    2.82619278517639096600e3,
    3.36093607810698293419e2,
]

YP1 = [
    1.26320474790178026440e9,
    -6.47355876379160291031e11,
    1.14509511541823727583e14,
    -8.12770255501325109621e15,
    2.02439475713594898196e17,
    -7.78877196265950026825e17,
]
YQ1 = [
    5.94301592346128195359e2,
    2.35564092943068577943e5,
    7.34811944459721705660e7,
    1.87601316108706159478e10,
    3.88231277496238566008e12,
    6.20557727146953693363e14,
    6.87141087355300489866e16,
    3.97270608116560655612e18,
]

Z1 = 1.46819706421238932572e1
Z2 = 4.92184563216946036703e1
PIO4 = 0.78539816339744830962  # pi/4
THPIO4 = 2.35619449019234492885  # 3*pi/4
SQ2OPI = 0.79788456080286535588  # sqrt(2/pi)

# polynomial coefficients for J0

PP0 = [
    7.96936729297347051624e-4,
    8.28352392107440799803e-2,
    1.23953371646414299388e0,
    5.44725003058768775090e0,
    8.74716500199817011941e0,
    5.30324038235394892183e0,
    9.99999999999999997821e-1,
]
PQ0 = [
    9.24408810558863637013e-4,
    8.56288474354474431428e-2,
    1.25352743901058953537e0,
    5.47097740330417105182e0,
    8.76190883237069594232e0,
    5.30605288235394617618e0,
    1.00000000000000000218e0,
]

QP0 = [
    -1.13663838898469149931e-2,
    -1.28252718670509318512e0,
    -1.95539544257735972385e1,
    -9.32060152123768231369e1,
    -1.77681167980488050595e2,
    -1.47077505154951170175e2,
    -5.14105326766599330220e1,
    -6.05014350600728481186e0,
]
QQ0 = [
    1.0,
    6.43178256118178023184e1,
    8.56430025976980587198e2,
    3.88240183605401609683e3,
    7.24046774195652478189e3,
    5.93072701187316984827e3,
    2.06209331660327847417e3,
    2.42005740240291393179e2,
]

YP0 = [
    1.55924367855235737965e4,
    -1.46639295903971606143e7,
    5.43526477051876500413e9,
    -9.82136065717911466409e11,
    8.75906394395366999549e13,
    -3.46628303384729719441e15,
    4.42733268572569800351e16,
    -1.84950800436986690637e16,
]
YQ0 = [
    1.04128353664259848412e3,
    6.26107330137134956842e5,
    2.68919633393814121987e8,
    8.64002487103935000337e10,
    2.02979612750105546709e13,
    3.17157752842975028269e15,
    2.50596256172653059228e17,
]

DR10 = 5.78318596294678452118e0
DR20 = 3.04712623436620863991e1

RP0 = [
    -4.79443220978201773821e9,
    1.95617491946556577543e12,
    -2.49248344360967716204e14,
    9.70862251047306323952e15,
]
RQ0 = [
    1.0,
    4.99563147152651017219e2,
    1.73785401676374683123e5,
    4.84409658339962045305e7,
    1.11855537045356834862e10,
    2.11277520115489217587e12,
    3.10518229857422583814e14,
    3.18121955943204943306e16,
    1.71086294081043136091e18,
]


def j0_small(x):
    """
    Implementation of J0 for x < 5
    """
    z = x * x
    # if x < 1.0e-5:
    #     return 1.0 - z/4.0
    p = (z - jnp.array(DR10)) * (z - jnp.array(DR20))
    p = p * jnp.polyval(jnp.array(RP0), z) / jnp.polyval(jnp.array(RQ0), z)
    return jnp.where(x < 1e-5, 1 - z / 4.0, p)


def j0_large(x):
    """
    Implementation of J0 for x >= 5
    """

    w = 5.0 / x
    q = 25.0 / (x * x)
    p = jnp.polyval(jnp.array(PP0), q) / jnp.polyval(jnp.array(PQ0), q)
    q = jnp.polyval(jnp.array(QP0), q) / jnp.polyval(jnp.array(QQ0), q)
    xn = x - PIO4
    p = p * jnp.cos(xn) - w * q * jnp.sin(xn)
    return p * SQ2OPI / jnp.sqrt(x)


@custom_jvp
def j0_(z):
    """
    Bessel function of the first kind of order zero and a real argument
    - using the implementation from CEPHES, translated to Jax, to match scipy to machine precision.
    Reference:
    Cephes Mathematical Library.
    Args:
        z: The sampling point(s) at which the Bessel function of the first kind are
        computed.
    Returns:
        An array of shape `x.shape` containing the values of the Bessel function
    """
    z = jnp.array(z)
    return jnp.where(jnp.abs(z) < 5.0, j0_small(jnp.abs(z)), j0_large(jnp.abs(z)))


@j0_.defjvp
def j0_jvp(primals, tangents):
    (m,) = primals
    (m_dot,) = tangents
    # Derivative of the complete elliptic integral of the first kind with respect to m
    return j0_(m), -1.0 * m_dot * j1_(m)


def j1_small(x):
    """
    Implementation of J1 for x < 5
    """
    z = x * x
    w = jnp.polyval(jnp.array(RP1), z) / jnp.polyval(jnp.array(RQ1), z)
    w = w * x * (z - jnp.array(Z1)) * (z - jnp.array(Z2))
    return w


def j1_large(x):
    """
    Implementation of J1 for x > 5
    """
    w = 5.0 / x
    z = w * w
    p = jnp.polyval(jnp.array(PP1), z) / jnp.polyval(jnp.array(PQ1), z)
    q = jnp.polyval(jnp.array(QP1), z) / jnp.polyval(jnp.array(QQ1), z)
    xn = x - THPIO4
    p = p * jnp.cos(xn) - w * q * jnp.sin(xn)
    return p * SQ2OPI / jnp.sqrt(x)


@custom_jvp
def j1_(z):
    """
    Bessel function of the first kind of order one and a real argument
    - using the implementation from CEPHES, translated to Jax, to match scipy to machine precision.
    Reference:
    Cephes mathematical library.
    Args:
        x: The sampling point(s) at which the Bessel function of the first kind are
        computed.
    Returns:
        An array of shape `x.shape` containing the values of the Bessel function
    """
    z = jnp.array(z)

    return jnp.sign(z) * jnp.where(
        jnp.abs(z) < 5.0, j1_small(jnp.abs(z)), j1_large(jnp.abs(z))
    )


@j1_.defjvp
def j1_jvp(primals, tangents):
    (m,) = primals
    (m_dot,) = tangents
    primal_out = j1_(m)
    tangent_out = j0_(m) * m_dot - j1_(m) * m_dot / m
    # tangent_out = lax.cond(m == 0, lambda _: 0.0, lambda _: tangent_out - j1(m) * m_dot / m, operand=None)
    return primal_out, tangent_out


j0 = jit(vmap(j0_, in_axes=(0,)))
j1 = jit(vmap(j1_, in_axes=(0,)))


def j2_(x):
    return (2 / x) * j1_(x) - j0_(x)


j2 = jit(vmap(j2_, in_axes=(0,)))


def j1p5_(x):
    return jnp.sqrt(2 / (jnp.pi * x)) * (jnp.sin(x) / x - jnp.cos(x))


j1p5 = jit(vmap(j1p5_, in_axes=(0,)))



import os
os.environ["JAX_PLATFORMS"] = "cpu"

import jax.numpy as jnp
from jax import config

config.update("jax_enable_x64", True)

from microjax.utils import (
    first_nonzero,
    last_nonzero,
    first_zero,
    min_zero_avoiding,
    max_zero_avoiding,
    mean_zero_avoiding,
    sparse_argsort,
    match_points,
    trapz_zero_avoiding,
)


def test_first_last_and_zero_helpers():
    x = jnp.array([0.0, 0.0, 3.0, 0.0, 5.0])
    assert int(first_nonzero(x)) == 2
    assert int(last_nonzero(x)) == 4
    assert int(first_zero(x)) == 0


def test_min_max_zero_avoiding():
    x = jnp.array([0.0, -2.0, 0.0, 1.0, 3.0])
    assert float(min_zero_avoiding(x)) == -2.0
    assert float(max_zero_avoiding(x)) == 3.0

    # all zeros -> mean_zero_avoiding returns 0.0
    x0 = jnp.zeros(5)
    assert float(mean_zero_avoiding(x0)) == 0.0

    # mixture with zeros -> ignore zeros in mean
    x1 = jnp.array([0.0, 2.0, 0.0, 4.0])
    assert float(mean_zero_avoiding(x1)) == 3.0


def test_sparse_argsort():
    a = jnp.array([0.0, 2.0, 0.0, -1.0, 3.0])
    # nan-safe argsort should order by non-zero values only
    idx = list(map(int, sparse_argsort(a)))
    # Extract non-zero portion in sorted order
    non_zero_sorted = [a[i] for i in idx if float(a[i]) != 0.0]
    assert non_zero_sorted == [-1.0, 2.0, 3.0]


def test_match_points_simple():
    # a and b are close but shuffled
    a = jnp.array([0.0 + 0.0j, 1.0 + 0.0j, -1.0 + 0.0j])
    b = jnp.array([1.0 + 0.0j, -1.0 + 0.0j, 0.0 + 0.0j])
    perm = list(map(int, match_points(a, b)))
    # perm should map a[i] to closest element index in b
    mapped = b[jnp.array(perm)]
    assert jnp.allclose(mapped, a)


def test_trapz_zero_avoiding_tail_removal():
    # Integrate a simple linear function and remove the last trapezoid explicitly
    x = jnp.linspace(0.0, 1.0, 6)  # 0,0.2,...,1.0
    y = 2.0 * x  # integral over [0,1] is 1.0

    # Full integral
    full_I = trapz_zero_avoiding(y, x, tail_idx=len(x) - 2)  # should drop last slice
    # Compute expected: integral over [0,0.8]
    # For linear y=2x, integral is x^2 from 0 to 0.8 -> 0.64
    assert jnp.isclose(full_I, 0.64, atol=1e-12)



"""Limb-darkening profiles and their JAX-friendly derivatives.

Currently implements a linear limb-darkening law with a custom JVP so that
gradients remain informative near the limb.
"""

import jax
import jax.numpy as jnp
from jax import jit
from functools import partial
from jax import custom_jvp
from typing import Union

Array = jnp.ndarray

#@partial(jit, static_argnames=("u1"))
@custom_jvp
def Is_limb_1st(d: Union[float, Array], u1: float = 0.0) -> Union[float, Array]:
    """Linear limb-darkening intensity profile, normalized.

    Implements: ``I(d) = I0 * [1 - u1 * (1 - mu)]`` with ``mu = sqrt(1-d^2)``
    and ``I0 = 3/(pi*(3-u1))`` so that the disk-integrated flux equals 1.

    Parameters
    - d: float/array – Radial distance normalized by ``rho`` (0 at center, 1 at limb).
    - u1: float – Linear limb-darkening coefficient in [0, 1].

    Returns
    - I: float/array – Normalized intensity; zeroed for ``d >= 1``.
    """
    mu = jnp.sqrt(1.0 - d**2)
    I0 = 3.0 / jnp.pi / (3.0 - u1)
    I  = I0 * (1.0 - u1 * (1.0 - mu))
    return jnp.where(d < 1.0, I, 0.0)

@Is_limb_1st.defjvp
def Is_limb_1st_jvp(primals, tangents):
    d, u1 = primals
    d_dot, u1_dot = tangents

    fac = 100.0
    mu = jnp.sqrt(jnp.maximum(1e-3, 1.0 - d**2))
    I0 = 3.0 / jnp.pi / (3.0 - u1)
    dI0_du1 = -3.0 / jnp.pi / (3.0 - u1)**2
    prof = 1.0 - u1 * (1.0 - mu)
    dprof_du1 = - (1.0 - mu)
    dprof_dmu = u1
    dmu_dd = -d / mu
    z = 1.0 - d
    sigmoid = jax.nn.sigmoid(fac * z)
    dsig_dz = fac * sigmoid * (1.0 - sigmoid)
    dz_dd = -1.0

    primal_out = I0 * prof * sigmoid
    tangent_out = u1_dot * (dI0_du1 * prof * sigmoid + I0 * dprof_du1 * sigmoid) \
      + d_dot * (I0 * dprof_dmu * dmu_dd * sigmoid + I0 * prof * dsig_dz * dz_dd)
    return primal_out, tangent_out



from microjax.fastlens.special import gamma
from microjax.fastlens.special import digamma_
from scipy.special import gamma as gamma_s
from scipy.special import digamma as digamma_s
import numpy as np
import jax.numpy as jnp
from jax import jit

from jax import config

config.update("jax_enable_x64", True)


def test_gamma():
    """
    This function tests the gamma function by comparing the results of the gamma function from microjax and scipy.
    The function generates random complex numbers and compares the results of the gamma function from microjax and scipy.
    The function passes if the results are close within a tolerance of 1e-14.
    """
    random_complex = np.random.uniform(
        low=-100, high=100, size=1000
    ) + 1.0j * np.random.uniform(low=-100, high=100, size=1000)
    random_complex_jax = jnp.array(random_complex)
    gamma_jax = gamma(random_complex_jax)
    gamma_scipy = gamma_s(random_complex)
    atol = 1e-14

    assert jnp.allclose(gamma_jax, gamma_scipy, atol=atol)


def test_digamma():
    """
    This function tests the digamma function by comparing the results of the digamma function from microjax and scipy.
    The function generates random complex numbers and compares the results of the digamma function from microjax and scipy.
    The function passes if the results are close within a tolerance of 1e-14.
    """
    random_complex = np.random.uniform(
        low=-100, high=100, size=1000
    ) + 1.0j * np.random.uniform(low=-100, high=100, size=1000)
    random_complex_jax = jnp.array(random_complex)
    digamma = jit(digamma_)
    digamma_jax = digamma(random_complex_jax)
    digamma_scipy = digamma_s(random_complex)
    atol = 1e-14

    assert jnp.allclose(digamma_jax, digamma_scipy, atol=atol)


if __name__ == "__main__":
    test_gamma()
    test_digamma()
    print("All tests passed!")



from microjax.fastlens.special import j0, j1, j2, j1p5
from scipy.special import jv as jv_s
from scipy.special import j0 as j0_s
from scipy.special import j1 as j1_s
import numpy as np
import jax.numpy as jnp
from jax import jit

from jax import config

config.update("jax_enable_x64", True)

def get_random_real():
    random_real = np.random.uniform(low=-1000, high=1000, size=1000)  
    random_real_jax = jnp.array(random_real)
    return random_real_jax

def test_j0():
    random_real_jax = get_random_real()
    j0_jax   = j0(random_real_jax)
    j0_scipy = j0_s(random_real_jax) 
    atol = 1e-14

    assert jnp.allclose(j0_jax, j0_scipy, atol=atol)
    
    
def test_j1():
    random_real_jax = get_random_real()    
    j1_jax   = j1(random_real_jax)
    j1_scipy = j1_s(random_real_jax) 
    atol = 1e-14

    assert jnp.allclose(j1_jax, j1_scipy, atol=atol)
    
def test_j2():
    random_real_jax = get_random_real()    
    j2_jax   = j2(random_real_jax)
    j2_scipy = jv_s(2.0,random_real_jax) 
    atol = 1e-14
    
    assert jnp.allclose(j2_jax, j2_scipy, atol=atol)

def test_j1p5():    
    random_real = np.random.uniform(low=0, high=1000, size=1000)  
    random_real_jax = jnp.array(random_real)
    j1p5_jax   = j1p5(random_real_jax)
    j1p5_scipy = jv_s(1.5,random_real_jax) 
    atol = 1e-14

    assert jnp.allclose(j1p5_jax, j1p5_scipy, atol=atol)



if __name__ == "__main__":
    test_j0()
    test_j1()
    test_j2()
    test_j1p5()
    print("All tests passed!")



import numpy as np
import jax.numpy as jnp
import pytest
from tests.utils.gpu import gpu_tests_enabled, has_cuda, is_a100

from microjax.inverse_ray.lightcurve import mag_triple


def make_trajectory(u0, tE, t0, alpha, n=24, span=2.0):
    t = t0 + jnp.linspace(-span * tE, span * tE, n)
    tau = (t - t0) / tE
    y1 = -u0 * jnp.sin(alpha) + tau * jnp.cos(alpha)
    y2 = u0 * jnp.cos(alpha) + tau * jnp.sin(alpha)
    w_points = jnp.array(y1 + 1j * y2, dtype=complex)
    return w_points


@pytest.mark.gpu
def test_triple_shapes_and_finiteness():
    if not (gpu_tests_enabled() and has_cuda() and is_a100()):
        pytest.skip("GPU/A100 not available or MICROJAX_GPU_TESTS not enabled")
    s, q, q3 = 0.9, 0.2, 0.1
    rho = 0.01
    alpha = jnp.deg2rad(20.0)
    w = make_trajectory(u0=0.05, tE=10.0, t0=0.0, alpha=alpha, n=24, span=1.0)
    mags = mag_triple(
        w,
        rho,
        s=s,
        q=q,
        q3=q3,
        r_resolution=64,
        th_resolution=64,
        Nlimb=64,
        bins_r=16,
        bins_th=32,
        margin_r=0.5,
        margin_th=0.5,
        MAX_FULL_CALLS=200,
        chunk_size=24,
    )
    arr = np.array(mags)
    assert arr.shape == (w.shape[0],)
    assert np.all(np.isfinite(arr))



import numpy as np
import jax.numpy as jnp
from jax import config

config.update("jax_enable_x64", True)

from microjax.point_source import (
    lens_eq,
    lens_eq_det_jac,
    _images_point_source,
    _images_point_source_sequential,
    mag_point_source,
    critical_and_caustic_curves,
)

import pytest


def test_binary_images_satisfy_lens_equation_and_mask_true():
    params = {"s": 1.2, "q": 0.3}
    a = 0.5 * params["s"]
    e1 = params["q"] / (1.0 + params["q"]) 
    w = 0.37 - 0.19j

    # Compute images in mid-point frame, lens_eq expects 'a' and 'e1'
    z, mask = _images_point_source(w, nlenses=2, a=a, e1=e1)
    res = np.array(lens_eq(np.array(z), nlenses=2, a=a, e1=e1) - np.array(w))
    ok = np.array(mask)
    assert ok.any()
    assert np.all(np.abs(res[ok]) < 1e-6)


def test_binary_magnification_matches_sum_of_inverse_det():
    params = {"s": 1.2, "q": 0.3}
    w = 0.37 - 0.19j
    # mag_point_source takes mid-point coord in COM internally; we mirror its logic here
    a = 0.5 * params["s"]
    e1 = params["q"] / (1.0 + params["q"]) 
    x_cm = a * (1.0 - params["q"]) / (1.0 + params["q"]) 
    w_mid = w - x_cm

    z, mask = _images_point_source(w_mid, nlenses=2, a=a, e1=e1)
    det = lens_eq_det_jac(z, nlenses=2, a=a, e1=e1)
    A_expected = (1.0 / jnp.abs(det)) * mask
    A_expected = np.sum(np.array(A_expected))

    A_num = float(np.array(mag_point_source(jnp.array(w), nlenses=2, **params)))
    assert np.isclose(A_num, A_expected, rtol=1e-10, atol=1e-10)

def test_critical_caustic_mapping_binary():
    params = {"s": 1.1, "q": 0.5}
    a = 0.5 * params["s"]
    e1 = params["q"] / (1.0 + params["q"]) 
    x_cm = a * (1.0 - params["q"]) / (1.0 + params["q"]) 
    z_cr, z_ca = critical_and_caustic_curves(npts=64, nlenses=2, **params)
    # Undo the COM shift to test mapping property in mid-point coords
    z_cr_mid = z_cr - x_cm
    z_ca_mid = z_ca - x_cm
    mapped = lens_eq(z_cr_mid, nlenses=2, a=a, e1=e1)
    assert np.allclose(np.array(mapped), np.array(z_ca_mid), rtol=1e-6, atol=1e-6)


def test_triple_images_shape_and_mask_reasonable():
    params = {"s": 0.9, "q": 0.3, "q3": 0.2, "r3": 0.4, "psi": 0.3}
    a = 0.5 * params["s"]
    e1 = params["q"] / (1.0 + params["q"] + params["q3"]) 
    e2 = 1.0 / (1.0 + params["q"] + params["q3"]) 
    w = -0.1 + 0.2j
    z, m = _images_point_source(w, nlenses=3, a=a, r3=params["r3"], psi=params["psi"], e1=e1, e2=e2)
    # For a triple lens, up to 10 images are possible; check non-zero mask entries
    assert np.array(z).ndim == 1
    assert np.sum(np.array(m)) >= 1


def test_triple_magnification_finite_far_field():
    params = {"s": 0.9, "q": 0.3, "q3": 0.2, "r3": 0.4, "psi": 0.3}
    w = 100.0 + 0.0j
    A = float(np.array(mag_point_source(jnp.array(w), nlenses=3, **params)))
    # Should approach 1 for very large |w|
    assert np.isclose(A, 1.0, rtol=1e-8, atol=1e-8)


def test_invalid_nlenses_raises():
    w = 0.0 + 0.0j
    try:
        _ = mag_point_source(w, nlenses=4)
        raised = False
    except ValueError:
        raised = True
    assert raised


@pytest.mark.parametrize("s", [0.1, 0.5, 2.0, 10.0])
@pytest.mark.parametrize("q", [1e-6, 1e-3, 0.1, 1.0])
def test_binary_far_field_magnification_over_param_grid(s, q):
    params = {"s": s, "q": q}
    # Large |w| should give A ~ 1
    w = jnp.array(100.0 + 0.0j)
    A = float(np.array(mag_point_source(w, nlenses=2, **params)))
    assert np.isclose(A, 1.0, rtol=1e-8, atol=1e-8)


@pytest.mark.parametrize("s", [0.1, 0.5, 2.0, 10.0])
@pytest.mark.parametrize("q", [1e-6, 1e-3, 0.1, 1.0])
def test_binary_mag_matches_det_over_param_grid(s, q):
    # Check A = sum 1/|det J| for valid images across parameter space
    params = {"s": s, "q": q}
    a = 0.5 * s
    e1 = q / (1.0 + q)
    x_cm = a * (1.0 - q) / (1.0 + q)
    w = 0.37 - 0.19j
    w_mid = w - x_cm

    z, mask = _images_point_source(w_mid, nlenses=2, a=a, e1=e1)
    det = lens_eq_det_jac(z, nlenses=2, a=a, e1=e1)
    A_expected = np.sum(np.array((1.0 / jnp.abs(det)) * mask))

    A_num = float(np.array(mag_point_source(jnp.array(w), nlenses=2, **params)))
    # Allow a modest tolerance due to numerical differences across wide params
    assert np.isfinite(A_num)
    assert np.isclose(A_num, A_expected, rtol=1e-8, atol=1e-10)



import jax
from jax import lax, vmap, jit
import numpy as np
import jax.numpy as jnp
from microjax.point_source import _images_point_source
from microjax.inverse_ray.merge_area import calc_source_limb, define_regions
from microjax.point_source import critical_and_caustic_curves
import matplotlib.pyplot as plt
plt.rcParams["figure.figsize"] = [6,6.0]
#plt.rcParams['text.usetex'] = True
import seaborn as sns
sns.set_theme(font="serif", font_scale=1.2,style="ticks")
jax.config.update("jax_enable_x64", True)
from functools import partial

w_center = jnp.complex128(-0.0510682 +0.1j)
#w_center = jnp.complex128(0.20732831+0.23517111j)
s = 1.1
q = 0.1
a = 0.5 * s
e1 = q / (1.0 + q)
_params = {"a": a, "e1": e1, "q": q, "s": s}
rho = 0.05

nlenses = 2
Nlimb = 200
margin_r  = 0.5
margin_th = 1.0
bins_r = 50
bins_th= 120

critical_curves, caustic_curves = critical_and_caustic_curves(nlenses=2, npts=500, s=s, q=q) 
image_limb, mask_limb = calc_source_limb(w_center, rho, Nlimb, **_params)
image_limb = image_limb.ravel()
mask_limb = mask_limb.ravel()
r_scan, th_scan = define_regions(image_limb, mask_limb, rho, bins_r=bins_r, bins_th=bins_th, 
                                 margin_r=margin_r, margin_th=margin_th, nlenses=nlenses)
for i, (r, th) in enumerate(zip(r_scan, th_scan)):
    print(i, r, th)

from microjax.point_source import lens_eq
shifted = 0.5 * s * (1 - q) / (1 + q)
w_center_shifted = w_center - shifted
r_resolution  = 50
th_resolution = 50
r_grid_norm = jnp.linspace(0, 1, r_resolution, endpoint=True)
th_grid_norm = jnp.linspace(0, 1, th_resolution, endpoint=True)
@jit
def plot(r_range, th_range):
    r_values = r_grid_norm * (r_range[1] - r_range[0]) + r_range[0]
    th_values = th_grid_norm * (th_range[1] - th_range[0]) + th_range[0]
    r_mesh, th_mesh = jnp.meshgrid(r_values, th_values, indexing='ij')
    z_grid = r_mesh * (jnp.cos(th_mesh) + 1j * jnp.sin(th_mesh))
    image_mesh = lens_eq(z_grid - shifted, **_params)
    distances = jnp.abs(image_mesh - w_center_shifted)
    in_source = (distances - rho < 0.0)
    return z_grid.real, z_grid.imag, in_source
vmap_plot = vmap(plot, in_axes=(0, 0))
x_grids, y_grids, in_sources = vmap_plot(r_scan, th_scan)

from mpl_toolkits.axes_grid1.inset_locator import inset_axes, mark_inset

fig = plt.figure(figsize=(6,6))
ax = plt.axes()
for x_grid, y_grid, in_source in zip(x_grids, y_grids, in_sources):
    ax.scatter(x_grid.ravel(), y_grid.ravel(), c='lightgray', s=0.5, zorder=0)
    ax.scatter(x_grid[in_source].ravel(), y_grid[in_source].ravel(), c='orange', s=0.5, zorder=1)

w_limb = w_center + jnp.array(rho * jnp.exp(1.0j * jnp.linspace(0.0, 2*jnp.pi, Nlimb)), dtype=complex)
ax.scatter(w_limb.real, w_limb.imag, color="blue", s=1, label="source limb")
critical_curves, caustic_curves = critical_and_caustic_curves(nlenses=2, npts=500, s=s, q=q) 
ax.scatter(critical_curves.ravel().real, critical_curves.ravel().imag, 
            marker=".", color="green", s=3, label="critical curve")
ax.scatter(caustic_curves.ravel().real, caustic_curves.ravel().imag, 
            marker=".", color="crimson", s=3, label="caustic")
ax.plot(-q/(1+q) * s, 0 , "o",c="k")
ax.plot((1.0)/(1+q) * s, 0 ,"o",c="k")
ax.scatter(image_limb[mask_limb].real, image_limb[mask_limb].imag, s=1, color="purple", label="image limb")
ax.scatter(0, 0, marker="x", color="k")
ax.grid(ls="--")
#ax.scatter(image_limb[~mask_limb].real, image_limb[~mask_limb].imag, s=1, color="lightblue")
ax.axis("equal")
ax.legend(loc="lower right", fontsize=9)

axins = inset_axes(ax, width="25%", height="25%", loc="upper right")
for x_grid, y_grid, in_source in zip(x_grids, y_grids, in_sources):
    axins.scatter(x_grid.ravel(), y_grid.ravel(), c='lightgray', s=0.5, zorder=0)
    axins.scatter(x_grid[in_source].ravel(), y_grid[in_source].ravel(), c='orange', s=0.5, zorder=1)
axins.scatter(image_limb[mask_limb].real, image_limb[mask_limb].imag, s=1, color="purple")
axins.scatter(w_limb.real, w_limb.imag, color="blue", s=1)
x0, y0 = 1.17, -0.04
dx = 0.025
x1, x2 = x0-dx, x0+dx  
y1, y2 = y0-dx, y0+dx  
axins.set_xlim(x1, x2)
axins.set_ylim(y1, y2)
ax.set_title(rf"$N_{{\rm limb}} = {Nlimb:d}$, $N_r={r_resolution:d}$, $N_\theta={th_resolution:d}$")
ax.set_xlabel("$x$ [$R_{\\rm E}$]")
ax.set_ylabel("$y$ [$R_{\\rm E}$]")
plt.savefig("example/visualize-icrs/visualize_example.png", bbox_inches="tight", dpi=300)
plt.show()